/*
 * Filename: FORMAT.CPP
 *
 * Description: Format WinSock 2 structures for output
 *
 * Edit History:
 *  08/31/96  rcq  created to enhance DT_DLL output
 */

#include "nowarn.h"  /* turn off benign warnings */
#ifndef _WINSOCKAPI_
#define _WINSOCKAPI_   /* Prevent inclusion of winsock.h in windows.h */
#endif
#include <windows.h>
#include "nowarn.h"  /* some warnings may have been turned back on */
#include <ws2spi.h>

#include "handlers.h"
#include "dt_dll.h"
#include "dt.h"

#define BYTES2DISPLAY 64
u_int nDataLimit = BYTES2DISPLAY;

extern char *fd_events[];

char *byte_order[] = {
	"BIGENDIAN",               // 0x0000
	"LITTLEENDIAN"             // 0x0001
};

char *guarantees [] = {
    "BestEffortService"
    "ControlledLoadService",
    "PredictiveService",
    "GuaranteedDelayService",
    "GuaranteedService"
};

//
// The comparators
//
char *comparators[] = {
    "COMP_EQUAL",
    "COMP_NOTLESS"
};

/*
 * Function: FormatProtocolInfo()
 *
  typedef struct _WSAPROTOCOL_INFOA {
    DWORD dwServiceFlags1;
    DWORD dwServiceFlags2;
    DWORD dwServiceFlags3;
    DWORD dwServiceFlags4;
    DWORD dwProviderFlags;
    GUID ProviderId;
    DWORD dwCatalogEntryId;
    WSAPROTOCOLCHAIN ProtocolChain;
    int iVersion;
    int iAddressFamily;
    int iMaxSockAddr;
    int iMinSockAddr;
    int iSocketType;
    int iProtocol;
    int iProtocolMaxOffset;
    int iNetworkByteOrder;
    int iSecurityScheme;
    DWORD dwMessageSize;
    DWORD dwProviderReserved;
    CHAR   szProtocol[257];
  } WSAPROTOCOL_INFOA, FAR * LPWSAPROTOCOL_INFOA;
 * Description: Format contents of Protocol Info structure for output
 */
INT FormatProtocolInfo (LPWSAPROTOCOL_INFO lpProtocolInfo, char *Buffer, INT off) {
	INT i;

	off += wsprintf(Buffer + off, 
		"ProtocolInfo\r\n\t\t.dwServiceFlags1=%#lx\r\n\t\t",
		lpProtocolInfo->dwServiceFlags1);
	off += wsprintf(Buffer + off, 
		".dwServiceFlags2=%#lx\r\n\t\t",
		lpProtocolInfo->dwServiceFlags2);
	off += wsprintf(Buffer + off, 
		".dwServiceFlags3=%#lx\r\n\t\t",
		lpProtocolInfo->dwServiceFlags3);
	off += wsprintf(Buffer + off, 
		".dwServiceFlags4=%#lx\r\n\t\t",
		lpProtocolInfo->dwServiceFlags4);
	off += wsprintf(Buffer + off, 
		".dwProviderFlags=%#lx\r\n\t\t",
		lpProtocolInfo->dwProviderFlags);
	off += wsprintf(Buffer + off,".ProviderId\r\n\t\t\t");
	off = FormatGUID (&(lpProtocolInfo->ProviderId), Buffer, off, "\t\t\t");
	off += wsprintf(Buffer + off, 
		"\t\t.dwCatalogEntryId=%lu\r\n\t\t",
		lpProtocolInfo->dwCatalogEntryId);

	off += wsprintf(Buffer + off,				// Protocol Chain
		".ProtocolChain.ChainLen=%lu\r\n\t\t",
		lpProtocolInfo->ProtocolChain.ChainLen);
    for (i=0; i<lpProtocolInfo->ProtocolChain.ChainLen;i++) {
		off += wsprintf(Buffer + off, "\t");
		off += wsprintf(Buffer + off, 
			".ProtocolChain.ChainEntries[%d]=%lu\r\n\t\t",
			i,lpProtocolInfo->ProtocolChain.ChainLen);
	}
	off += wsprintf(Buffer + off, 
		".iVersion=%d\r\n\t\t",
		lpProtocolInfo->iVersion);
	off += wsprintf(Buffer + off, 
		".iAddressFamily=%d\r\n\t\t",
		lpProtocolInfo->iAddressFamily);
	off += wsprintf(Buffer + off, 
		".iMaxSockAddr=%s\r\n\t\t",
		(lpProtocolInfo->iMinSockAddr<2) ?
		byte_order[lpProtocolInfo->iMinSockAddr] : 
		"<value out of range>");
	off += wsprintf(Buffer + off, 
		".iSocketType=%d\r\n\t\t",
		lpProtocolInfo->iSocketType);
	off += wsprintf(Buffer + off, 
		".iProtocol=%d\r\n\t\t",
		lpProtocolInfo->iProtocol);
    off += wsprintf(Buffer + off, 
		".iProtocolMaxOffset=%d\r\n\t\t",
		lpProtocolInfo->iProtocolMaxOffset);
	off += wsprintf(Buffer + off, 
		".iNetworkByteOrder=%d\r\n\t\t",
		lpProtocolInfo->iNetworkByteOrder);
	off += wsprintf(Buffer + off, 
		".iSecurityScheme=%d\r\n\t\t",
		lpProtocolInfo->iSecurityScheme);
	off += wsprintf(Buffer + off, 
		".dwMessageSize=%lu\r\n\t\t",
		lpProtocolInfo->dwMessageSize);
	off += wsprintf(Buffer + off, 
		".dwProviderReserved=%lu\r\n\t\t",
		lpProtocolInfo->dwProviderReserved);
	off += wsprintf (Buffer + off, ".szProtocol: %s\r\n",
		lpProtocolInfo->szProtocol);

	return (off);
} // end FormatProtocolInfo


/*-------------------------------------------------------------------*
 * WSA Overlapped structure
 *
  typedef struct _WSAOVERLAPPED {
    DWORD    Internal;		// reserved
    DWORD    InternalHigh;	// reserved
    DWORD    Offset;		// reserved
    DWORD    OffsetHigh;	// reserved
    WSAEVENT hEvent;		// ** Event Handle **
  } WSAOVERLAPPED, FAR * LPWSAOVERLAPPED;
 *
 *  Used by functions:
 *   Overlapped Completion Routine
 *   WSAGetOverlappedResult()
 *   WSAIoctl()
 *   WSARecv()
 *   WSARecvFrom()
 *   WSASend()
 *   WSASendTo()
 -------------------------------------------------------------------*/
INT FormatWSAOverlapped (LPWSAOVERLAPPED lpOverlapped, char *Buffer, INT off) {
	// Just display the single field that matters to us:
	off += wsprintf (Buffer + off, 
		"lpOverlapped.hEvent: %d\r\n",
		lpOverlapped->hEvent);
	return (off);
} // end FormatWSAOverlapped()

  
/*-------------------------------------------------------------------*
 * WSA Buffer structure
 *
  typedef struct _WSABUF {
    u_long      len;     // the length of the buffer 
    char FAR *  buf;     // the pointer to the buffer
  } WSABUF, FAR * LPWSABUF;
 *
 *  Used by functions:
 *   WSAGetQOSByName()
 *   WSAJoinLeaf()
 *   WSARecv()
 *   WSARecvDisconnect()
 *   WSARecvFrom()
 *   WSASend()
 *   WSASendDisconnect()
 *   WSASendTo()
 -------------------------------------------------------------------*/
 INT FormatWSABuf (LPWSABUF lpWSABuf, char *Buffer, INT off, BOOL bDoData) {
 	 u_int i;
	 u_long len = lpWSABuf->len;

	 off += wsprintf (Buffer + off, 
	 	"WSABuf.len: %d (0x%lx)\r\n\t", len, len);
	 if (lpWSABuf->buf) {
		 if (bDoData) {
			off += wsprintf (Buffer + off, 
				 "WSABuf.buf contents is:\r\n\t\t");
			for (i=0; ((i<len) && (i<nDataLimit)); i++) {
				if (i && !(i % 16))
					off += wsprintf (Buffer + off, "\r\n\t\t");
				off += wsprintf (Buffer + off,
					"0x%02x ", *((lpWSABuf->buf)+i));
			}
			off += wsprintf (Buffer + off, "\r\n");
		 } else {
			off += wsprintf (Buffer + off, "WSBuf.buf looks ok\r\n");
		 }
	 } else {
		 off += wsprintf (Buffer + off, "WSABuf.buf is NULL\r\n");
	 }
	return (off);
 } // end FormatWSABuf()
  
/*-------------------------------------------------------------------*
 *  QOS Flow specifications (unidirectional)
 *
  typedef struct _flowspec {
    int32        TokenRate;              // In Bytes/sec 

    int32        TokenBucketSize;        // In Bytes 
    int32        PeakBandwidth;          // In Bytes/sec 
    int32        Latency;                // In microseconds 
    int32        DelayVariation;         // In microseconds 
    GUARANTEE    LevelOfGuarantee;       // Guaranteed, Predictive 
                                         //   or Best Effort       
    int32        CostOfCall;             // Reserved for future use, 
                                         //   must be set to 0 now   
    int32        NetworkAvailability;    // read-only:         
                                         //   1 if accessible, 
                                         //   0 if not         
  } FLOWSPEC, FAR * LPFLOWSPEC;
 *
 *  Used by functions:
 *   none - all reference QOS structure
 -------------------------------------------------------------------*/
INT FormatFlowSpec (LPFLOWSPEC lpFlowSpec, char *Buffer, INT off) {
    off += wsprintf (Buffer + off, "TokenRate: %d (0x%lx)\r\n",
		lpFlowSpec->TokenRate, lpFlowSpec->TokenRate);
    off += wsprintf (Buffer + off, "TokenBucketSize: %d (0x%lx)\r\n",
		lpFlowSpec->TokenBucketSize, lpFlowSpec->TokenBucketSize);
    off += wsprintf (Buffer + off, "PeakBandwidth: %d (0x%lx)\r\n",
		lpFlowSpec->PeakBandwidth, lpFlowSpec->PeakBandwidth);
    off += wsprintf (Buffer + off, "Latency: %d (0x%lx)\r\n",
		lpFlowSpec->Latency, lpFlowSpec->Latency);
    off += wsprintf (Buffer + off, "DelayVariation: %d (0x%lx)\r\n",
		lpFlowSpec->DelayVariation, lpFlowSpec->DelayVariation);
    off += wsprintf (Buffer + off, "CostOfCall: %d (0x%lx)\r\n",
		lpFlowSpec->CostOfCall, lpFlowSpec->CostOfCall);
    off += wsprintf (Buffer + off, "LevelOfGuarantee: %s\r\n",
		guarantees [lpFlowSpec->LevelOfGuarantee]);
    off += wsprintf (Buffer + off, "NetworkAvailability: %s (0x%lx)\r\n",
		lpFlowSpec->NetworkAvailability ? "TRUE" : "FALSE",
		lpFlowSpec->NetworkAvailability);
	return (off);
} // end FormatFlowSpec()

/*-------------------------------------------------------------------*
 * Quality of Service
 *
  typedef struct _QualityOfService {
    FLOWSPEC      SendingFlowspec;       // the flow spec for data sending 
    FLOWSPEC      ReceivingFlowspec;     // the flow spec for data receiving 
    WSABUF        ProviderSpecific;      // additional provider specific stuff 
  } QOS, FAR * LPQOS;
 *
 *  Used by functions:
 *   WSAAccept Condition Procedure
 *   WSAConnect()
 *   WSAIoctl()
 *   WSAGetQOSByName()
 *   WSAJoinLeaf()
 -------------------------------------------------------------------**/
INT FormatQOS (LPQOS lpQOS, char *Buffer, INT off) {
	off += wsprintf (Buffer + off, "QOS.SendingFlowspec:\r\n");
	off = FormatFlowSpec (&(lpQOS->SendingFlowspec), Buffer, off);
	off += wsprintf (Buffer + off, "QOS.ReceivingFlowspec:\r\n");
	off = FormatFlowSpec (&(lpQOS->ReceivingFlowspec), Buffer, off);
	off += wsprintf (Buffer + off, "QOS.ProviderSpecific:\r\n");
	off = FormatWSABuf (&(lpQOS->ProviderSpecific), Buffer, off, TRUE);
	return (off);
} // end FormatQOS() 

/*-------------------------------------------------------------------*
 * WSA Network Events (from WSAEnumNetworkEvents())
 *
	typedef struct _WSANETWORKEVENTS {
       long lNetworkEvents;				// bit flags of FD Events that occurred
       int iErrorCode[FD_MAX_EVENTS];	// WSA error for each active event
	} WSANETWORKEVENTS, FAR * LPWSANETWORKEVENTS;
 *
 *  Used by functions:
 *   WSAEnumNetworkEvents()
 -------------------------------------------------------------------*/
INT FormatWSANetworkEvents (LPWSANETWORKEVENTS lpNetEvents, char *Buffer, INT off) {
	int i;
	
	off += wsprintf (Buffer + off, "lpNetEvents.lNetworkEvents: ");
	for (i=0; i<FD_MAX_EVENTS; i++) {
		if ((lpNetEvents->lNetworkEvents) & (1<<i)) {
			off += wsprintf(Buffer + off, "%s ", fd_events[i]); 
		}
	}
	off += wsprintf(Buffer + off, "\r\n\tlpNetEvents.iErrorCode: ");
	for (i=0; i<FD_MAX_EVENTS; i++) {
		if ((lpNetEvents->lNetworkEvents) & (1<<i)) {
			int nWSAErr = lpNetEvents->iErrorCode[i];
			off += wsprintf(Buffer + off, "%d (0x%lx)", nWSAErr, nWSAErr);
		}
	}
	return (off);
} // end FormatWSANetworkEvents()

/*-------------------------------------------------------------------*
 * Protocol Chain
 *
  typedef struct _WSAPROTOCOLCHAIN {
    int ChainLen;                           // the length of the chain,    
                                            // length = 0 means layered protocol,
                                            // length = 1 means base protocol, 
                                            // length > 1 means protocol chain 
    DWORD ChainEntries[MAX_PROTOCOL_CHAIN]; // a list of dwCatalogEntryIds 
  } WSAPROTOCOLCHAIN, FAR * LPWSAPROTOCOLCHAIN;
 *
 *  Used by functions:
 *   none - referenced in WSAProtocolInfo structure
 -------------------------------------------------------------------*/
INT FormatProtocolChain (LPWSAPROTOCOLCHAIN lpProtocolChain, char *Buffer, INT off) {
	return (off);
} // end FormatProtocolChain()

/*-------------------------------------------------------------------*
 * the GUID - Globally Unique Identifier
 *
	typedef struct _GUID
	{
		unsigned long  Data1;
		unsigned short Data2;
		unsigned short Data3;
		unsigned char  Data4[8];
	} GUID;
 *
 *  Used by structures
 *   WSAPROTOCOL_INFO structure
 *   WSAQUERYSET
 *   WSASERVICECLASSINFO
 *   WSANAMESPACEINFO
 *
 *  Used in functions
 *   WSARemoveServiceClass()
 *   WSAGetServiceClassInfo()
 *   WSAGetServiceClassNameByClassId()
 -------------------------------------------------------------------*/
INT FormatGUID (LPGUID lpGUID, char *Buffer, INT off, char *tab) {
	int i;
	off += wsprintf(Buffer + off, "Data1=%lu\r\n%s",
		lpGUID->Data1, tab);
	off += wsprintf(Buffer + off, "Data2=%lu\r\n%s",
		lpGUID->Data2, tab);
	off += wsprintf(Buffer + off, "Data3=%lu\r\n%s",
		lpGUID->Data3, tab);
	off += wsprintf(Buffer + off, "Data4=");
    for (i=0; i<8; i++) {
		off += wsprintf (Buffer + off,
			"0x%02x ", lpGUID->Data4[i]);
	}
	off += wsprintf (Buffer + off, "\r\n");
	return (off);
} // end FormatGUID()

/*-------------------------------------------------------------------*
 * the BLOB
 *
	typedef struct _BLOB {
		ULONG cbSize ;
	#ifdef MIDL_PASS
		[size_is(cbSize)] BYTE *pBlobData;
	#else    // MIDL_PASS
		BYTE *pBlobData ;
	#endif   // MIDL_PASS 
	} BLOB, *LPBLOB ;
	#endif
 *
 *  Used by functions:
 *    none - part of WSAQUERYSET structure
 -------------------------------------------------------------------*/
INT FormatBlob (LPBLOB lpBlob, char *Buffer, INT off, char *tab) {
	off += wsprintf (Buffer + off, "Blob.cbSize: %d (0x%lx)\r\n%s",
		lpBlob->cbSize, lpBlob->cbSize,tab);
	off += wsprintf (Buffer + off, "Blob.pBlobData: <bq note: not done>\r\n");
	return (off);
} // end FormatBlob()

 /*-------------------------------------------------------------------*
 * SockAddr Information
 *
 typedef struct _SOCKET_ADDRESS {
    LPSOCKADDR lpSockaddr ;
    INT iSockaddrLength ;
  } SOCKET_ADDRESS, *PSOCKET_ADDRESS, FAR * LPSOCKET_ADDRESS ;
 *
 *  Used by functions:
 *   none - part of CSADDR_INFO structure
 -------------------------------------------------------------------*/
INT FormatSocketAddress (LPSOCKET_ADDRESS lpSocketAddress, char *Buffer, INT off, char *tab) {
	if (lpSocketAddress->lpSockaddr) {
		if (!IsBadReadPtr(lpSocketAddress->lpSockaddr, sizeof(SOCKADDR))) {
			int nAF = lpSocketAddress->lpSockaddr->sa_family;
			if (nAF == AF_INET) {
				off += wsprintf (Buffer + off, "SocketAddress.lpSockaddr:\r\n%s",tab);
				off = FormatSockaddrIn ((sockaddr_in*)lpSocketAddress->lpSockaddr, Buffer, off);
				off += wsprintf (Buffer + off, tab);
			} else {
				off += wsprintf (Buffer + off, 
					"SocketAddress.lpSockaddr unknown address family: %d (0x%lx)\r\n%s",
					nAF, nAF, tab);
			}
		} else {
			off += wsprintf (Buffer + off,
				"SocketAddress.lpSockaddr is an INVALID POINTER\r\n%s",tab);
		}
	} else {
		off += wsprintf (Buffer + off,
			"SocketAddress.lpSockaddr is a NULL pointer\r\n%s",tab);
	}
	off += wsprintf (Buffer + off, 
		"SocketAddress.iSockaddrLength: %d (0x%lx)\r\n",
		lpSocketAddress->iSockaddrLength, lpSocketAddress->iSockaddrLength);
	return (off);
} // end FormatSocketAddress()

/*-------------------------------------------------------------------*
 * CSAddr Information
 * 
  typedef struct _CSADDR_INFO {
    SOCKET_ADDRESS LocalAddr ;
    SOCKET_ADDRESS RemoteAddr ;
    INT iSocketType ;
    INT iProtocol ;
  } CSADDR_INFO, *PCSADDR_INFO, FAR * LPCSADDR_INFO ;
 *
 *  Used by functions:
 *    none - part of WSAQuerySet structure
 -------------------------------------------------------------------*/
INT FormatCSAddrInfo (LPCSADDR_INFO lpCSAddrInfo, char *Buffer, INT off, char *tab) {
	off += wsprintf (Buffer + off, "CSAddrInfo.LocalAddr:\r\n%s",tab);
	off  = FormatSocketAddress (&(lpCSAddrInfo->LocalAddr), Buffer, off, tab);
	off += wsprintf (Buffer + off, "CSAddrInfo.RemoteAddr:\r\n%s",tab);
	off  = FormatSocketAddress (&(lpCSAddrInfo->RemoteAddr), Buffer, off, tab);
	off += wsprintf (Buffer + off, "CSAddrInfo.iSocketType: %d (0x%lx)\r\n%s",
		lpCSAddrInfo->iSocketType, lpCSAddrInfo->iSocketType,tab);
	off += wsprintf (Buffer + off, "CSAddrInfo.iProtocol: %d (0x%lx)\r\n",
		lpCSAddrInfo->iProtocol, lpCSAddrInfo->iProtocol);
	return (off);
} // end FormatCSAddrInfo()

/*-------------------------------------------------------------------*
 *  Address Family/Protocol Tuples
 *
  typedef struct _AFPROTOCOLS {
    INT iAddressFamily;
    INT iProtocol;
  } AFPROTOCOLS, *PAFPROTOCOLS, *LPAFPROTOCOLS;
 *
 *  Used by functions:
 *   none - part of WSAQuerySet structure
 -------------------------------------------------------------------*/
INT FormatAFProtocols (LPAFPROTOCOLS lpAFProts, char *Buffer, INT off, char *tab) {
	off += wsprintf (Buffer + off, 
		"AFProtocols.iAddressFamily: %d (0x%lx)\r\n%s",
		lpAFProts->iAddressFamily, lpAFProts->iAddressFamily,tab);
	off += wsprintf (Buffer + off, 
		"AFProtocols.iProtocol: %d (0x%lx)\r\n",
		lpAFProts->iProtocol, lpAFProts->iProtocol);
	return (off);
} // end FormatAFProtocols()

/*-------------------------------------------------------------------*
 * Service Installation/Removal Data Types.
 *
  typedef struct _WSANSClassInfoA {
    LPSTR   lpszName;
    DWORD   dwNameSpace;
    DWORD   dwValueType;
    DWORD   dwValueSize;
    LPVOID  lpValue;
  }WSANSCLASSINFOA, *PWSANSCLASSINFOA, *LPWSANSCLASSINFOA;
 *
 *  Used by functions:
 *   none - part of ServiceClassInfo structure
 -------------------------------------------------------------------*/
INT FormatNSClassInfo (LPWSANSCLASSINFO lpNSClass, char *Buffer, INT off, char *tab) {
	if (lpNSClass->lpszName) {
		off += wsprintf (Buffer + off,
			"NSClassInfo.lpszName: %s\r\n%s", lpNSClass->lpszName, tab);
	} else {
		off += wsprintf (Buffer + off,
			"NSClassInfo.lpszName is NULL\r\n%s",tab);
	}
	off += wsprintf (Buffer + off, 
		"NSClassInfo.dwNameSpace: %d (0x%lx)\r\n%s",
		lpNSClass->dwNameSpace, lpNSClass->dwNameSpace, tab);
	off += wsprintf (Buffer + off, 
		"NSClassInfo.dwValueType: %d (0x%lx)\r\n%s",
		lpNSClass->dwValueType, lpNSClass->dwValueType, tab);
	off += wsprintf (Buffer + off, 
		"NSClassInfo.dwValueSize: %d (0x%lx)\r\n%s",
		lpNSClass->dwValueSize, lpNSClass->dwValueSize, tab);
	if (lpNSClass->lpValue) {
		if (!IsBadReadPtr (lpNSClass->lpValue, sizeof(int))) {
			off += wsprintf (Buffer + off, 
				"NSClassInfo.lpValue: %d (0x%lx)\r\n",
				*((int*)lpNSClass->lpValue), *((int*)lpNSClass->lpValue));
		} else {
			off += wsprintf (Buffer + off,
				"NSClassInfo.lpValue is an INVALID POINTER\r\n");
		}
	} else {
		off += wsprintf (Buffer + off,
			"NSClassInfo.lpValue is NULL\r\n");
	}
	return (off);
} // end FormatNSClassInfo()

/*-------------------------------------------------------------------*
 * Service Installation/Removal Data Types.
 *
  typedef struct _WSANSClassInfoW {
    LPSTR   lpszName;
    DWORD   dwNameSpace;
    DWORD   dwValueType;
    DWORD   dwValueSize;
    LPVOID  lpValue;
  }WSANSCLASSINFOA, *PWSANSCLASSINFOA, *LPWSANSCLASSINFOA;
 *
 *  Used by functions:
 *   none - part of ServiceClassInfo structure
 -------------------------------------------------------------------*/
INT FormatNSClassInfoW (LPWSANSCLASSINFOW lpNSClass, char *Buffer, INT off, char *tab) {
	if (lpNSClass->lpszName) {
		off += wsprintf (Buffer + off,
			"NSClassInfo.lpszName: %s\r\n%s", lpNSClass->lpszName, tab);
	} else {
		off += wsprintf (Buffer + off,
			"NSClassInfo.lpszName is NULL\r\n%s",tab);
	}
	off += wsprintf (Buffer + off, 
		"NSClassInfo.dwNameSpace: %d (0x%lx)\r\n%s",
		lpNSClass->dwNameSpace, lpNSClass->dwNameSpace, tab);
	off += wsprintf (Buffer + off, 
		"NSClassInfo.dwValueType: %d (0x%lx)\r\n%s",
		lpNSClass->dwValueType, lpNSClass->dwValueType, tab);
	off += wsprintf (Buffer + off, 
		"NSClassInfo.dwValueSize: %d (0x%lx)\r\n%s",
		lpNSClass->dwValueSize, lpNSClass->dwValueSize, tab);
	if (lpNSClass->lpValue) {
		off += wsprintf (Buffer + off, 
			"NSClassInfo.lpValue: %d (0x%lx)\r\n",
			*((int*)lpNSClass->lpValue), *((int*)lpNSClass->lpValue));
	} else {
		off += wsprintf (Buffer + off,
			"NSClassInfo.lpValue is NULL\r\n");
	}
	return (off);
} // end FormatNSClassInfoW()

/*-------------------------------------------------------------------*
 * Service Class Information
 *
  typedef struct _WSAServiceClassInfoA {
    LPGUID              lpServiceClassId;
    LPSTR               lpszServiceClassName;
    DWORD               dwCount;
    LPWSANSCLASSINFOA   lpClassInfos;
  WSASERVICECLASSINFOA, *PWSASERVICECLASSINFOA, *LPWSASERVICECLASSINFOA;
 *
 *  Used by functions:
 *   WSAInstallServiceClass()
 *   WSAGetServiceClassInfo()
 -------------------------------------------------------------------*/
INT FormatServiceClassInfo (LPWSASERVICECLASSINFO lpServiceClass, char *Buffer, INT off, char *tab) {
	if (lpServiceClass->lpServiceClassId) {
		if (!IsBadReadPtr(lpServiceClass->lpServiceClassId, sizeof(GUID))) {
			off += wsprintf (Buffer + off,
				"ServiceClassInfo.lpServiceClassId: \r\n%s",tab);
			off = FormatGUID (lpServiceClass->lpServiceClassId, Buffer, off, "\t\t\t");
			off += wsprintf (Buffer + off, "%s", tab);
		} else {
			off += wsprintf (Buffer + off, 
				"ServiceClassInfo.lpServiceClassId is an INVALID POINTER\r\n%s",tab);
		}
	} else {
		off += wsprintf (Buffer + off,
			"ServiceClassInfo.lpServiceClassId is NULL\r\n%s",tab);
	}
	if (lpServiceClass->lpszServiceClassName) {
		off += wsprintf (Buffer + off,
			"ServiceClassInfo.lpszServiceClassName: %s\r\n%s",
			lpServiceClass->lpszServiceClassName, tab);
	} else {
		off += wsprintf (Buffer + off,
			"ServiceClassInfo.lpszServiceClassName is NULL\r\n%s", tab);
	}
	off += wsprintf (Buffer + off, "ServiceClassInfo.dwCount: %d (0x%lx)\r\n%s",
		lpServiceClass->dwCount, lpServiceClass->dwCount,tab);
	if (lpServiceClass->lpClassInfos) {
		if (!IsBadReadPtr(lpServiceClass->lpClassInfos, sizeof(GUID))) {
			off += wsprintf (Buffer + off, "ServiceClassInfo.lpClassInfos:\r\n%s",tab);
			off = FormatNSClassInfo (lpServiceClass->lpClassInfos, Buffer, off, "\t\t\t");
		} else {
			off += wsprintf (Buffer + off, 
				"ServiceClassInfo.lpClassInfos is an INVALID POINTER\r\n%s",tab);
		}
	} else {
		off += wsprintf (Buffer + off,
			"ServiceClassInfo.lpClassInfos is NULL\r\n%s",tab);
	}

	return (off);
} // end FormatServiceClassInfo()

/*-------------------------------------------------------------------*
 * Service Class Information
 *
  typedef struct _WSAServiceClassInfoW {
    LPGUID              lpServiceClassId;
    LPSTR               lpszServiceClassName;
    DWORD               dwCount;
    LPWSANSCLASSINFOW   lpClassInfos;
  WSASERVICECLASSINFOA, *PWSASERVICECLASSINFOA, *LPWSASERVICECLASSINFOA;
 *
 *  Used by functions:
 *   WSAInstallServiceClass()
 *   WSAGetServiceClassInfo()
 -------------------------------------------------------------------*/
INT FormatServiceClassInfoW (LPWSASERVICECLASSINFOW lpServiceClass, char *Buffer, INT off, char *tab) {
	if (lpServiceClass->lpServiceClassId) {
		if (!IsBadReadPtr(lpServiceClass->lpServiceClassId, sizeof(GUID))) {
			off += wsprintf (Buffer + off,
				"ServiceClassInfo.lpServiceClassId: \r\n%s",tab);
			off = FormatGUID (lpServiceClass->lpServiceClassId, Buffer, off, "\t\t\t");
			off += wsprintf (Buffer + off, "%s", tab);
		} else {
			off += wsprintf (Buffer + off, 
				"ServiceClassInfo.lpServiceClassId is an INVALID POINTER\r\n%s",tab);
		}
	} else {
		off += wsprintf (Buffer + off,
			"ServiceClassInfo.lpServiceClassId is NULL\r\n%s",tab);
	}
	if (lpServiceClass->lpszServiceClassName) {
		off += wsprintf (Buffer + off,
			"ServiceClassInfo.lpszServiceClassName: %s\r\n%s",
			lpServiceClass->lpszServiceClassName,tab);
	} else {
		off += wsprintf (Buffer + off,
			"ServiceClassInfo.lpszServiceClassName is NULL\r\n%s",tab);
	}
	off += wsprintf (Buffer + off, "ServiceClassInfo.dwCount: %d (0x%lx)\r\n%s",
		lpServiceClass->dwCount, lpServiceClass->dwCount,tab);
	off += wsprintf (Buffer + off, "ServiceClassInfo.lpClassInfos:\r\n%s",tab);
	off = FormatNSClassInfoW (lpServiceClass->lpClassInfos, Buffer, off, "\t\t\t");
	return (off);
} // end FormatServiceClassInfoW()

/*-------------------------------------------------------------------*
 * Name Space information
 *
  typedef struct _WSANAMESPACE_INFOA {
    GUID                NSProviderId;
    DWORD               dwNameSpace;
    BOOL                fActive;
    DWORD               dwVersion;
    LPSTR               lpszIdentifier;
  } WSANAMESPACE_INFOA, *PWSANAMESPACE_INFOA, *LPWSANAMESPACE_INFOA;
 *
 *  Used by functions:
 *   WSAEnumNameSpaceProviders()
 -------------------------------------------------------------------*/
INT FormatNameSpaceInfo (LPWSANAMESPACE_INFO lpNameSpace, char *Buffer, INT off, char *tab) {
	off += wsprintf (Buffer + off, "WSANameSpace_Info.NSProviderId:\r\n%s",tab);
	off =  FormatGUID (&(lpNameSpace->NSProviderId), Buffer, off, "/t/t/t");
	off += wsprintf (Buffer + off, 
		"WSANameSpace_Info.dwNameSpace: %d (0x%lx)\r\n%s,tab",
		lpNameSpace->dwNameSpace, lpNameSpace->dwNameSpace);
	off += wsprintf (Buffer + off,
		"WSANameSpace_Info.fActive: %s\r\n%s",
		lpNameSpace->fActive ? "TRUE" : "FALSE",tab);
	off += wsprintf (Buffer + off, 
		"WSANameSpace_Info.dwVersion: %d (0x%lx)\r\n%s",
		lpNameSpace->dwVersion, lpNameSpace->dwVersion,tab);
	off += wsprintf (Buffer + off, 
		"WSANameSpace_Info.lpszIdentifier: %s\r\n",
		lpNameSpace->lpszIdentifier);
	return (off);
} // end FormatNameSpaceInfo()

/*-------------------------------------------------------------------*
 * WSAVersion
 *
  typedef struct _WSAVersion {
    DWORD           dwVersion;
    WSAECOMPARATOR  ecHow;
  }WSAVERSION, *PWSAVERSION, *LPWSAVERSION;
 *
 *  Used by functions:
 *   none - part of Protocol Info structure
 *-------------------------------------------------------------------*/
INT FormatWSAVersion (LPWSAVERSION lpWSAVersion, char *Buffer, INT off, char *tab) {
	off += wsprintf (Buffer + off, "WSAVersion.dwVersion: %d (0x%lx)\r\n%s",
		lpWSAVersion->dwVersion, lpWSAVersion->dwVersion, tab);
	off += wsprintf (Buffer + off, "WSAVersion.ecHow: %s\r\n",
		comparators[lpWSAVersion->ecHow]);
	return (off);
} // end FormatWSAVersion()

/*-------------------------------------------------------------------*
 * Query Set
 *
  typedef struct _WSAQuerySetA {
    DWORD           dwSize;
    LPSTR           lpszServiceInstanceName;
    LPGUID          lpServiceClassId;
    LPWSAVERSION    lpVersion;
    LPSTR           lpszComment;
    DWORD           dwNameSpace;
    LPGUID          lpNSProviderId;
    LPSTR           lpszContext;
    DWORD           dwNumberOfProtocols;
    LPAFPROTOCOLS   lpafpProtocols;
    LPSTR           lpszQueryString;
    DWORD           dwNumberOfCsAddrs;
    LPCSADDR_INFO   lpcsaBuffer;
    DWORD           dwOutputFlags;
    LPBLOB          lpBlob;
  } WSAQUERYSETA, *PWSAQUERYSETA, *LPWSAQUERYSETA;
 *
 *  Used by functions:
 *   WSALookupServiceBegin()
 *   WSALookupServiceNext()
 *   WSASetService()
 -------------------------------------------------------------------*/
INT FormatQuerySet (LPWSAQUERYSET lpQuerySet, char *Buffer, INT off, char *tab) {
	off += wsprintf (Buffer + off, "WSAQuerySet.dwSize: %d (0x%lx)\r\n%s",
		 lpQuerySet->dwSize, lpQuerySet->dwSize, tab);
	if (lpQuerySet->lpszServiceInstanceName) {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpszServiceInstanceName: %s\r\n%s",
			lpQuerySet->lpszServiceInstanceName,tab);
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpszServiceInstanceName is NULL\r\n%s",tab);
	}
	if (lpQuerySet->lpServiceClassId) {
		if (!IsBadReadPtr(lpQuerySet->lpServiceClassId, sizeof(GUID))) {
			off += wsprintf (Buffer + off, "WSAQuerySet.lpServiceClassId: \r\n%s","\t\t\t");
			off = FormatGUID (lpQuerySet->lpServiceClassId, Buffer, off, "\t\t\t");
			off += wsprintf (Buffer + off, tab);
		} else {
			off += wsprintf (Buffer + off, 
				"WSAQuerySet.lpServiceClassId POINTER IS INVALID\r\n%s",tab);
		}
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpServiceClassId is NULL\r\n%s",tab);
	}
	if (lpQuerySet->lpVersion) {
		if (!IsBadReadPtr(lpQuerySet->lpVersion, sizeof(WSAVERSION))) {
			off += wsprintf (Buffer + off, "WSAQuerySet.lpVersion: \r\n%s","\t\t\t");
			off = FormatWSAVersion (lpQuerySet->lpVersion, Buffer, off, "\t\t\t");
			off += wsprintf (Buffer + off, tab);
		} else {
			off += wsprintf (Buffer + off, 
				"WSAQuerySet.lpVersion POINTER IS INVALID\r\n%s",tab);
		}
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpVersion is NULL\r\n%s",tab);
	}
	if (lpQuerySet->lpszComment) {
		off += wsprintf (Buffer + off, "WSAQuerySet.lpszComment: %s\r\n%s",
			lpQuerySet->lpszComment,tab);
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpszComment is NULL\r\n%s",tab);
	}
	off += wsprintf (Buffer + off,
		"WSAQuerySet.dwNameSpace: %d (0x%lx)\r\n%s",
		lpQuerySet->dwNameSpace, lpQuerySet->dwNameSpace,tab);
	if (lpQuerySet->lpNSProviderId) {
		if (!IsBadReadPtr(lpQuerySet->lpNSProviderId, sizeof(GUID))) {
			off += wsprintf (Buffer + off, "WSAQuerySet.lpNSProviderId: \r\n%s","\t\t\t");
			off = FormatGUID (lpQuerySet->lpNSProviderId, Buffer, off, "\t\t\t");
			off += wsprintf (Buffer + off, tab);
		} else {
			off += wsprintf (Buffer + off, 
				"WSAQuerySet.lpNSProviderId POINTER IS INVALID\r\n%s",tab);
		}
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpNSProviderId is NULL\r\n%s",tab);
	}
	if (lpQuerySet->lpszContext) {
		off += wsprintf (Buffer + off, "WSAQuerySet.lpszContext: %s\r\n%s",
			lpQuerySet->lpszContext,tab);
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpszContext is NULL\r\n%s",tab);
	}
	off += wsprintf (Buffer + off,
		"WSAQuerySet.dwNumberOfProtocols: %d (0x%lx)\r\n%s",
		lpQuerySet->dwNumberOfProtocols, lpQuerySet->dwNumberOfProtocols,tab);
	if (lpQuerySet->lpafpProtocols) {
		//
		// bq note: need a for loop for dwNumberOfProtocols
		//
		if (!IsBadReadPtr(lpQuerySet->lpafpProtocols, sizeof(AFPROTOCOLS))) {
			off += wsprintf (Buffer + off, "WSAQuerySet.lpafpProtocols: \r\n%s","\t\t\t");
			off = FormatAFProtocols (lpQuerySet->lpafpProtocols, Buffer, off, "\t\t\t");
			off += wsprintf (Buffer + off, tab);
		} else {
			off += wsprintf (Buffer + off, 
				"WSAQuerySet.lpafpProtocols POINTER IS INVALID\r\n%s",tab);
		}
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpafpProtocols is NULL\r\n%s",tab);
	}
	if (lpQuerySet->lpszQueryString) {
		off += wsprintf (Buffer + off, "WSAQuerySet.lpszQueryString: %s\r\n%s",
			lpQuerySet->lpszQueryString,tab);
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpszQueryString is NULL\r\n%s",tab);
	}
	off += wsprintf (Buffer + off,
		"WSAQuerySet.dwNumberOfCsAddrs: %d (0x%lx)\r\n%s",
		lpQuerySet->dwNumberOfCsAddrs, lpQuerySet->dwNumberOfCsAddrs,tab);
	if (lpQuerySet->lpcsaBuffer) {
		//
		// bq note: need a for loop for dwNumberOfCSAddrs
		//
		if (!IsBadReadPtr(lpQuerySet->lpcsaBuffer, sizeof(CSADDR_INFO))) {
			off += wsprintf (Buffer + off, "WSAQuerySet.lpcsaBuffer: \r\n%s","\t\t\t");
			off = FormatCSAddrInfo (lpQuerySet->lpcsaBuffer, Buffer, off, "\t\t\t");
			off += wsprintf (Buffer + off, tab);
		} else {
			off += wsprintf (Buffer + off, 
				"WSAQuerySet.lpcsaBuffer POINTER IS INVALID\r\n%s",tab);
		}
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpcsaBuffer is NULL\r\n%s",tab);
	}
	off += wsprintf (Buffer + off,
		"WSAQuerySet.dwOutputFlags: %d (0x%lx)\r\n%s",
		lpQuerySet->dwOutputFlags, lpQuerySet->dwOutputFlags,tab);
	if (lpQuerySet->lpBlob) {
		if (!IsBadReadPtr(lpQuerySet->lpBlob, sizeof(BLOB))) {
			off += wsprintf (Buffer + off, "WSAQuerySet.lpBlob: \r\n%s","\t\t\t");
			off = FormatBlob (lpQuerySet->lpBlob, Buffer, off, "\t\t\t");
		} else {
			off += wsprintf (Buffer + off, 
				"WSAQuerySet.lpBlob POINTER IS INVALID\r\n%s",tab);
		}
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpBlob is NULL\r\n");
	}
	return (off);
} // end FormatQuerySet()

/*-------------------------------------------------------------------*
 * Query Set
 *
  typedef struct _WSAQuerySetW {
    DWORD           dwSize;
    LPSTR           lpszServiceInstanceName;
    LPGUID          lpServiceClassId;
    LPWSAVERSION    lpVersion;
    LPSTR           lpszComment;
    DWORD           dwNameSpace;
    LPGUID          lpNSProviderId;
    LPSTR           lpszContext;
    DWORD           dwNumberOfProtocols;
    LPAFPROTOCOLS   lpafpProtocols;
    LPSTR           lpszQueryString;
    DWORD           dwNumberOfCsAddrs;
    LPCSADDR_INFO   lpcsaBuffer;
    DWORD           dwOutputFlags;
    LPBLOB          lpBlob;
  } WSAQUERYSETA, *PWSAQUERYSETA, *LPWSAQUERYSETA;
 *
 *  Used by functions:
 *   WSALookupServiceBegin()
 *   WSALookupServiceNext()
 *   WSASetService()
 -------------------------------------------------------------------*/
INT FormatQuerySetW (LPWSAQUERYSETW lpQuerySet, char *Buffer, INT off, char *tab) {
	off += wsprintf (Buffer + off, "WSAQuerySet.dwSize: %d (0x%lx)\r\n%s",
		 lpQuerySet->dwSize, lpQuerySet->dwSize, tab);
	if (lpQuerySet->lpszServiceInstanceName) {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpszServiceInstanceName: %s\r\n%s",
			lpQuerySet->lpszServiceInstanceName, tab);
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpszServiceInstanceName is NULL\r\n%s", tab);
	}
	if (lpQuerySet->lpServiceClassId) {
		if (!IsBadReadPtr(lpQuerySet->lpServiceClassId, sizeof(GUID))) {
			off += wsprintf (Buffer + off, "WSAQuerySet.lpServiceClassId: \r\n%s","\t\t\t");
			off = FormatGUID (lpQuerySet->lpServiceClassId, Buffer, off, "\t\t\t");
			off += wsprintf (Buffer + off, tab);
		} else {
			off += wsprintf (Buffer + off, 
				"WSAQuerySet.lpServiceClassId POINTER IS INVALID\r\n%s",tab);
		}
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpServiceClassId is NULL\r\n%s",tab);
	}
	if (lpQuerySet->lpVersion) {
		if (!IsBadReadPtr(lpQuerySet->lpVersion, sizeof(WSAVERSION))) {
			off += wsprintf (Buffer + off, "WSAQuerySet.lpVersion: \r\n%s","\t\t\t");
			off = FormatWSAVersion (lpQuerySet->lpVersion, Buffer, off, "\t\t\t");
			off += wsprintf (Buffer + off, tab);
		} else {
			off += wsprintf (Buffer + off, 
				"WSAQuerySet.lpVersion POINTER IS INVALID\r\n%s",tab);
		}
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpVersion is NULL\r\n%s",tab);
	}
	if (lpQuerySet->lpszComment) {
		off += wsprintf (Buffer + off, "WSAQuerySet.lpszComment: %s\r\n%s",
			lpQuerySet->lpszComment, tab);
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpszComment is NULL\r\n%s",tab);
	}
	off += wsprintf (Buffer + off,
		"WSAQuerySet.dwNameSpace: %d (0x%lx)\r\n%s",
		lpQuerySet->dwNameSpace, lpQuerySet->dwNameSpace,tab);
	if (lpQuerySet->lpNSProviderId) {
		if (!IsBadReadPtr(lpQuerySet->lpNSProviderId, sizeof(GUID))) {
			off += wsprintf (Buffer + off, "WSAQuerySet.lpNSProviderId: \r\n%s","\t\t\t");
			off = FormatGUID (lpQuerySet->lpNSProviderId, Buffer, off, "\t\t\t");
			off += wsprintf (Buffer + off, tab);
		} else {
			off += wsprintf (Buffer + off, 
				"WSAQuerySet.lpNSProviderId POINTER IS INVALID\r\n%s",tab);
		}
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpNSProviderId is NULL\r\n%s",tab);
	}
	if (lpQuerySet->lpszContext) {
		off += wsprintf (Buffer + off, "WSAQuerySet.lpszContext: %s\r\n%s",
			lpQuerySet->lpszContext,tab);
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpszContext is NULL\r\n%s",tab);
	}
	off += wsprintf (Buffer + off,
		"WSAQuerySet.dwNumberOfProtocols: %d (0x%lx)\r\n%s",
		lpQuerySet->dwNumberOfProtocols, lpQuerySet->dwNumberOfProtocols, tab);
	if (lpQuerySet->lpafpProtocols) {
		//
		// bq note: need a for loop for dwNumberOfProtocols
		//
		if (!IsBadReadPtr(lpQuerySet->lpafpProtocols, sizeof(AFPROTOCOLS))) {
			off += wsprintf (Buffer + off, "WSAQuerySet.lpafpProtocols: \r\n%s","\t\t\t");
			off = FormatAFProtocols (lpQuerySet->lpafpProtocols, Buffer, off, "\t\t\t");
			off += wsprintf (Buffer + off, tab);
		} else {
			off += wsprintf (Buffer + off, 
				"WSAQuerySet.lpafpProtocols POINTER IS INVALID\r\n%s",tab);
		}
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpafpProtocols is NULL\r\n%s",tab);
	}
	if (lpQuerySet->lpszQueryString) {
		off += wsprintf (Buffer + off, "WSAQuerySet.lpszQueryString: %s\r\n%s",
			lpQuerySet->lpszQueryString,tab);
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpszQueryString is NULL\r\n%s",tab);
	}
	off += wsprintf (Buffer + off,
		"WSAQuerySet.dwNumberOfCsAddrs: %d (0x%lx)\r\n%s",
		lpQuerySet->dwNumberOfCsAddrs, lpQuerySet->dwNumberOfCsAddrs,tab);
	if (lpQuerySet->lpcsaBuffer) {
		//
		// bq note: need a for loop for dwNumberOfCSAddrs
		//
		if (!IsBadReadPtr(lpQuerySet->lpcsaBuffer, sizeof(CSADDR_INFO))) {
			off += wsprintf (Buffer + off, "WSAQuerySet.lpcsaBuffer: \r\n%s","\t\t\t");
			off = FormatCSAddrInfo (lpQuerySet->lpcsaBuffer, Buffer, off, "\t\t\t");
			off += wsprintf (Buffer + off, tab);
		} else {
			off += wsprintf (Buffer + off, 
				"WSAQuerySet.lpcsaBuffer POINTER IS INVALID\r\n%s",tab);
		}
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpcsaBuffer is NULL\r\n%s",tab);
	}
	off += wsprintf (Buffer + off,
		"WSAQuerySet.dwOutputFlags: %d (0x%lx)\r\n%s",
		lpQuerySet->dwOutputFlags, lpQuerySet->dwOutputFlags,tab);
	if (lpQuerySet->lpBlob) {
		off += wsprintf (Buffer + off, "WSAQuerySet.lpBlob:\r\n%s","\t\t\t");
		off = FormatBlob (lpQuerySet->lpBlob, Buffer, off, "\t\t\t");
	} else {
		off += wsprintf (Buffer + off,
			"WSAQuerySet.lpBlob is NULL\r\n");
	}
	return (off);
} // end FormatQuerySetW()

/*-------------------------------------------------------------------*
 * WSPData
 *
  typedef struct WSPData {
    WORD         wVersion;
    WORD         wHighVersion;
    WCHAR        szDescription[WSPDESCRIPTION_LEN+1];
  } WSPDATA, FAR * LPWSPDATA;
 *
 *  Used by functions:
 *   WSPStartup()
 -------------------------------------------------------------------*/
INT FormatWSPData (LPWSPDATA lpWSPData, char *Buffer, INT off) {
	off += wsprintf(Buffer + off, ".wVersion: %d\r\n\t\t",
		lpWSPData->wVersion); 
	off += wsprintf(Buffer + off, ".wHighVersion: %d\r\n\t\t",
		lpWSPData->wHighVersion); 
	off += wsprintf(Buffer + off, ".szDescription: %s\r\n",
		lpWSPData->szDescription);
	return (off);
} // end FormatWSPData()

/*-------------------------------------------------------------------*
 * WSPThreadID
 *
  typedef struct _WSATHREADID {
    HANDLE ThreadHandle;
    DWORD Reserved;
  } WSATHREADID, FAR * LPWSATHREADID;
 *
 *  Used by functions:
 *   ??
 -------------------------------------------------------------------*/
INT FormatThreadID (LPWSATHREADID lpThreadID, char *Buffer, INT off) {
	return (off);
} // end FormatThreadID()

/*-------------------------------------------------------------------*
 * SP Procedure Table
 *
  typedef struct _WSPPROC_TABLE {

    LPWSPACCEPT              lpWSPAccept;
    LPWSPADDRESSTOSTRING     lpWSPAddressToString;
    LPWSPASYNCSELECT         lpWSPAsyncSelect;
    LPWSPBIND                lpWSPBind;
    LPWSPCANCELBLOCKINGCALL  lpWSPCancelBlockingCall;
    LPWSPCLEANUP             lpWSPCleanup;
    LPWSPCLOSESOCKET         lpWSPCloseSocket;
    LPWSPCONNECT             lpWSPConnect;
    LPWSPDUPLICATESOCKET     lpWSPDuplicateSocket;
    LPWSPENUMNETWORKEVENTS   lpWSPEnumNetworkEvents;
    LPWSPEVENTSELECT         lpWSPEventSelect;
    LPWSPGETOVERLAPPEDRESULT lpWSPGetOverlappedResult;
    LPWSPGETPEERNAME         lpWSPGetPeerName;
    LPWSPGETSOCKNAME         lpWSPGetSockName;
    LPWSPGETSOCKOPT          lpWSPGetSockOpt;
    LPWSPGETQOSBYNAME        lpWSPGetQOSByName;
    LPWSPIOCTL               lpWSPIoctl;
    LPWSPJOINLEAF            lpWSPJoinLeaf;
    LPWSPLISTEN              lpWSPListen;
    LPWSPRECV                lpWSPRecv;
    LPWSPRECVDISCONNECT      lpWSPRecvDisconnect;
    LPWSPRECVFROM            lpWSPRecvFrom;
    LPWSPSELECT              lpWSPSelect;
    LPWSPSEND                lpWSPSend;
    LPWSPSENDDISCONNECT      lpWSPSendDisconnect;
    LPWSPSENDTO              lpWSPSendTo;
    LPWSPSETSOCKOPT          lpWSPSetSockOpt;
    LPWSPSHUTDOWN            lpWSPShutdown;
    LPWSPSOCKET              lpWSPSocket;
    LPWSPSTRINGTOADDRESS     lpWSPStringToAddress;

  } WSPPROC_TABLE, FAR * LPWSPPROC_TABLE;
* 
*  Used by Functions:
*   WSPStartup()
*/
INT FormatProcTable (LPWSPPROC_TABLE lpProcTable, char *Buffer, INT off) {
	off += wsprintf(Buffer + off,
		"lpWSPAccept: 0x%08lx\r\n\t\t", 
		lpProcTable->lpWSPAccept);
	off += wsprintf(Buffer + off,
	    "lpWSPAddressToString: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPAddressToString);
	off += wsprintf(Buffer + off,
	    "lpWSPAsyncSelect: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPAsyncSelect);
	off += wsprintf(Buffer + off,
	    "lpWSPBind: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPBind);
	off += wsprintf(Buffer + off,
	    "lpWSPCancelBlockingCall: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPCancelBlockingCall);
	off += wsprintf(Buffer + off,
	    "lpWSPCleanup: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPCleanup);
	off += wsprintf(Buffer + off,
	    "lpWSPCloseSocket: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPCloseSocket);
	off += wsprintf(Buffer + off,
	    "lpWSPConnect: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPConnect);
	off += wsprintf(Buffer + off,
	    "lpWSPDuplicateSocket: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPDuplicateSocket);
	off += wsprintf(Buffer + off,
	    "lpWSPEnumNetworkEvents: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPEnumNetworkEvents);
	off += wsprintf(Buffer + off,
	    "lpWSPEventSelect: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPEventSelect);
	off += wsprintf(Buffer + off,
	    "lpWSPGetOverlappedResult: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPGetOverlappedResult);
	off += wsprintf(Buffer + off,
	    "lpWSPGetPeerName: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPGetPeerName);
	off += wsprintf(Buffer + off,
	    "lpWSPGetSockName: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPGetSockName);
	off += wsprintf(Buffer + off,
	    "lpWSPGetSockOpt: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPGetSockOpt);
	off += wsprintf(Buffer + off,
	    "lpWSPGetQOSByName: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPGetQOSByName);
	off += wsprintf(Buffer + off,
	    "lpWSPIoctl: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPIoctl);
	off += wsprintf(Buffer + off,
	    "lpWSPJoinLeaf: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPJoinLeaf);
	off += wsprintf(Buffer + off,
	    "lpWSPListen: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPListen);
	off += wsprintf(Buffer + off,
	    "lpWSPRecv: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPRecv);
	off += wsprintf(Buffer + off,
	    "lpWSPRecvDisconnect: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPRecvDisconnect);
	off += wsprintf(Buffer + off,
	    "lpWSPRecvFrom: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPRecvFrom);
	off += wsprintf(Buffer + off,
	    "lpWSPSelect: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPSelect);
	off += wsprintf(Buffer + off,
	    "lpWSPSend: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPSend);
	off += wsprintf(Buffer + off,
	    "lpWSPSendDisconnect: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPSendDisconnect);
	off += wsprintf(Buffer + off,
	    "lpWSPSendTo: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPSendTo);
	off += wsprintf(Buffer + off,
	    "lpWSPSetSockOpt: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPSetSockOpt);
	off += wsprintf(Buffer + off,
	    "lpWSPShutdown: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPShutdown);
	off += wsprintf(Buffer + off,
	    "lpWSPSocket: 0x%08lx\r\n\t\t",
		lpProcTable->lpWSPSocket);
	off += wsprintf(Buffer + off,
	    "lpWSPStringToAddress: 0x%08lx\r\n",
		lpProcTable->lpWSPStringToAddress);

	return (off);
} // end FormatProcTable()

/*-------------------------------------------------------------------*
 * SP Upcall Table
 *
typedef struct _WSPUPCALLTABLE {

    LPWPUCLOSEEVENT               lpWPUCloseEvent;
    LPWPUCLOSESOCKETHANDLE        lpWPUCloseSocketHandle;
    LPWPUCREATEEVENT              lpWPUCreateEvent;
    LPWPUCREATESOCKETHANDLE       lpWPUCreateSocketHandle;
    LPWPUFDISSET                  lpWPUFDIsSet;
    LPWPUGETPROVIDERPATH          lpWPUGetProviderPath;
    LPWPUMODIFYIFSHANDLE          lpWPUModifyIFSHandle;
    LPWPUPOSTMESSAGE              lpWPUPostMessage;
    LPWPUQUERYBLOCKINGCALLBACK    lpWPUQueryBlockingCallback;
    LPWPUQUERYSOCKETHANDLECONTEXT lpWPUQuerySocketHandleContext;
    LPWPUQUEUEAPC                 lpWPUQueueApc;
    LPWPURESETEVENT               lpWPUResetEvent;
    LPWPUSETEVENT                 lpWPUSetEvent;
    LPWPUOPENCURRENTTHREAD        lpWPUOpenCurrentThread;
    LPWPUCLOSETHREAD              lpWPUCloseThread;

} WSPUPCALLTABLE, FAR * LPWSPUPCALLTABLE;
* 
*  Used by Functions:
*   WSPStartup()
*/
INT FormatUpcallTable (LPWSPUPCALLTABLE lpUpcallTable, char *Buffer, INT off) {
	off += wsprintf(Buffer + off,
		"lpWPUCloseEvent: 0x%08lx\r\n\t\t", 
		lpUpcallTable->lpWPUCloseEvent);
	off += wsprintf(Buffer + off,
	    "lpWPUCloseSocketHandle: 0x%08lx\r\n\t\t",
		lpUpcallTable->lpWPUCloseSocketHandle);
	off += wsprintf(Buffer + off,
	    "lpWPUCreateEvent: 0x%08lx\r\n\t\t",
		lpUpcallTable->lpWPUCreateEvent);
	off += wsprintf(Buffer + off,
	    "lpWPUCreateSocketHandle: 0x%08lx\r\n\t\t",
		lpUpcallTable->lpWPUCreateSocketHandle);
	off += wsprintf(Buffer + off,
	    "lplpWPUFDIsSet: 0x%08lx\r\n\t\t",
		lpUpcallTable->lpWPUFDIsSet);
	off += wsprintf(Buffer + off,
	    "lpWPUGetProviderPath: 0x%08lx\r\n\t\t",
		lpUpcallTable->lpWPUGetProviderPath);
	off += wsprintf(Buffer + off,
	    "lpWPUModifyIFSHandle: 0x%08lx\r\n\t\t",
		lpUpcallTable->lpWPUModifyIFSHandle);
	off += wsprintf(Buffer + off,
	    "lpWPUPostMessage: 0x%08lx\r\n\t\t",
		lpUpcallTable->lpWPUPostMessage);
	off += wsprintf(Buffer + off,
	    "lpWPUQueryBlockingCallback: 0x%08lx\r\n\t\t",
		lpUpcallTable->lpWPUQueryBlockingCallback);
	off += wsprintf(Buffer + off,
	    "lpWPUQuerySocketHandleContext: 0x%08lx\r\n\t\t",
		lpUpcallTable->lpWPUQuerySocketHandleContext);
	off += wsprintf(Buffer + off,
	    "lpWPUQueueApc: 0x%08lx\r\n\t\t",
		lpUpcallTable->lpWPUQueueApc);
	off += wsprintf(Buffer + off,
	    "lpWPUResetEvent: 0x%08lx\r\n\t\t",
		lpUpcallTable->lpWPUResetEvent);
	off += wsprintf(Buffer + off,
	    "lpWPUSetEvent: 0x%08lx\r\n\t\t",
		lpUpcallTable->lpWPUSetEvent);
	off += wsprintf(Buffer + off,
	    "lpWPUOpenCurrentThread: 0x%08lx\r\n\t\t",
		lpUpcallTable->lpWPUOpenCurrentThread);
	off += wsprintf(Buffer + off,
	    "lpWPUCloseThread: 0x%08lx\r\n",
		lpUpcallTable->lpWPUCloseThread);
	return (off);
} // end FormatUpcallTable()

/*-------------------- BSD compatible structures -------------------*/ 

/*-------------------------------------------------------------------*
 * Timeval structure
 *
	struct timeval {
        long    tv_sec;         //seconds 
        long    tv_usec;        // and microseconds 
	};
*/
INT FormatTimeval (LPTIMEVAL lpTimeval, char *Buffer, INT off) {
	off += wsprintf (Buffer + off, "timeval.tv_sec:  %d (0x%lx)\r\n",
		lpTimeval->tv_sec, lpTimeval->tv_sec);
	off += wsprintf (Buffer + off, "timeval.tv_usec: %d (0x%lx)\r\n",
		lpTimeval->tv_usec, lpTimeval->tv_usec);
	return (off);
} // end FormatTimeval()

/*-------------------------------------------------------------------*
 * Hostent structure
 *
	struct  hostent {
        char    FAR * h_name;           // official name of host 
        char    FAR * FAR * h_aliases;  // alias list 
        short   h_addrtype;             // host address type 
        short   h_length;               // length of address 
        char    FAR * FAR * h_addr_list;// list of addresses 
	#define h_addr  h_addr_list[0]      // address, for backward compat
	};
*/
INT FormatHostent (LPHOSTENT lpHostent, char *Buffer, INT off) {
	int i;
	off += wsprintf (Buffer + off, "hostent.h_name: %s\r\n",
		lpHostent->h_name);
	for (i=0; (lpHostent->h_aliases)+i; i++) {
		off += wsprintf (Buffer + off, "hostent.h_aliases: %s\r\n",
			(lpHostent->h_aliases)+i);
	}
	off += wsprintf (Buffer + off, "hostent.h_addrtype: %d (0x%lx)\r\n",
		lpHostent->h_addrtype, lpHostent->h_addrtype);
	off += wsprintf (Buffer + off, "hostent.h_length: %d (0x%lx)\r\n",
		lpHostent->h_length, lpHostent->h_length);
	for (i=0; (lpHostent->h_addr_list)+i; i++) {
		off += wsprintf (Buffer + off, 
			"hostent.h_addr_list: %d (0x%lx)\r\n",
			(lpHostent->h_addr_list)+i, (lpHostent->h_addr_list)+i);
	}
	return (off);
} // end FormatHostent()

/*-------------------------------------------------------------------* 
 * Netent structure
 *
	struct  netent {
        char    FAR * n_name;          // official name of net 
        char    FAR * FAR * n_aliases; // alias list 
        short   n_addrtype;            // net address type 
        u_long  n_net;                 // network # 
	};
*/
INT FormatNetent (struct netent FAR *lpNetent, char *Buffer, INT off) {
	return (off);
} // end FormatNetent()

/*-------------------------------------------------------------------* 
 * Servent structure
 *
	struct  servent {
        char    FAR * s_name;           // official service name
        char    FAR * FAR * s_aliases;  // alias list
        short   s_port;                 // port #
        char    FAR * s_proto;          // protocol to use
	};
*/
INT FormatServent (LPSERVENT lpServent, char *Buffer, INT off) {
	return (off);
} // end FormatServent()


/*-------------------------------------------------------------------* 
 * Protoent structure
 *
	struct  protoent {
        char    FAR * p_name;           // official protocol name
        char    FAR * FAR * p_aliases;  // alias list
        short   p_proto;                // protocol #
	};
*/
INT FormatProtoent (LPPROTOENT lpProtoent, char *Buffer, INT off) {
	return (off);
} // end FormatProtoent()

/*-------------------------------------------------------------------* 
 * In_Addr structure
 *
	struct in_addr {
        union {
                struct { u_char s_b1,s_b2,s_b3,s_b4; } S_un_b;
                struct { u_short s_w1,s_w2; } S_un_w;
                u_long S_addr;
        } S_un;

*/
INT FormatInAddr (LPIN_ADDR lpInAddr, char *Buffer, INT off) {
	off += wsprintf (Buffer+off, 
		"in_addr: %d.%d.%d.%d\r\n",
		lpInAddr->S_un.S_un_b.s_b1,
		lpInAddr->S_un.S_un_b.s_b2,
		lpInAddr->S_un.S_un_b.s_b3,
		lpInAddr->S_un.S_un_b.s_b4);
	return (off);
} // end FormatInAddr()

/*-------------------------------------------------------------------* 
 * SockAddr_In structure
 *
	struct sockaddr_in {
        short   sin_family;
        u_short sin_port;
        struct  in_addr sin_addr;
        char    sin_zero[8];
	};
*/
INT FormatSockaddrIn (LPSOCKADDR_IN lpSockAddrIn, char *Buffer, INT off) {
	off += wsprintf (Buffer + off, 
		"sockaddr_in.sin_family: %d, sin_port: %d, sin_addr: %d.%d.%d.%d (0x%lx)\r\n",
		lpSockAddrIn->sin_family, 
//		(((lpSockAddrIn->sin_port & 0xFF)<<8) | ((lpSockAddrIn->sin_port & 0xFF00)>>8)),
		lpSockAddrIn->sin_port,
		lpSockAddrIn->sin_addr.S_un.S_un_b.s_b1,
		lpSockAddrIn->sin_addr.S_un.S_un_b.s_b2,
		lpSockAddrIn->sin_addr.S_un.S_un_b.s_b3,
		lpSockAddrIn->sin_addr.S_un.S_un_b.s_b4,
		lpSockAddrIn->sin_addr.S_un.S_addr);
	return (off);
} // end FormatSockAddrIn()

/*-------------------------------------------------------------------* 
 * WSAData structure
 *
	typedef struct WSAData {
        WORD                    wVersion;
        WORD                    wHighVersion;
        char                    szDescription[257];
        char                    szSystemStatus[129];
        unsigned short          iMaxSockets;
        unsigned short          iMaxUdpDg;
        char FAR *              lpVendorInfo;
	} WSADATA, FAR * LPWSADATA;
*/
INT FormatWSAData (LPWSADATA lpWSAData, char *Buffer, INT off) {
	off += wsprintf(Buffer + off, 
		"lpWSAData:\r\n\t\t.wVersion: %#lx  wHighVersion: %#lx\r\n\t\t",
		lpWSAData->wVersion, lpWSAData->wHighVersion);
	if (lpWSAData->szDescription[0]) {
		off += wsprintf(Buffer + off,
			"szDescription: %s\r\n\t\t",
			lpWSAData->szDescription);
	} else {
		off += wsprintf(Buffer + off,"szDescription is NULL");
	}
	if (lpWSAData->szSystemStatus[0]) {
		off += wsprintf(Buffer + off,
			"szSystemStatus: %s\r\n\t\t",
			lpWSAData->szSystemStatus);
	} else {
		off += wsprintf(Buffer + off,"szSystemStatus is NULL");
	}
	off += wsprintf(Buffer + off,
		"iMaxSockets: %d  iMaxUdpDg: %d\r\n\t\t",
		lpWSAData->iMaxSockets, lpWSAData->iMaxUdpDg);
	off += wsprintf(Buffer + off,
		"lpVendorInfo pointer is %s\r\n",
		lpWSAData->lpVendorInfo ? "non-NULL" : "NULL");

	return (off);
} // end FormatWSAData()

/*-------------------------------------------------------------------* 
 * Linger structure
 *
	struct  linger {
        u_short l_onoff;          // option on/off
        u_short l_linger;         // linger time 
	};
 */
INT FormatLinger (LPLINGER lpLinger, char *Buffer, INT off) {
	off += wsprintf (Buffer + off,
		"Linger.l_onoff: %d, l_linger: %d\r\n",
		lpLinger->l_onoff, lpLinger->l_linger);
	return (off);
} // end FormatLinger()

