/*
 * Commands for ioctlsocket(),  taken from the BSD file fcntl.h.
 *
 *
 * Ioctl's have the command encoded in the lower word,
 * and the size of any in or out parameters in the upper
 * word.  The high 2 bits of the upper word are used
 * to encode the in/out status of the parameter; for now
 * we restrict parameters to at most 128 bytes.
 */
char ioctl_ioc[] = {
	"IOC_VOID",		/* 0x2000000 */
	"IOC_OUT",		/* 0x4000000 */
	"IOC_IN",		/* 0x8000000 */
	"IOC_INOUT"		/* 0xC000000 */
};

char ioctl_fio[] = {
	"FIONASYNC",	/* 0xd */
	"FIONBIO",		/* 0xe */
	"FIONREAD"		/* 0xf */
};

char ioctl_sio[] = {
	"SIOCSHIWAT",	/* 0 */	
	"SIOCGHIWAT",	/* 1 */
	"SIOCSLOWAT",	/* 2 */
	"SIOCGLOWAT",	/* 3 */
	/0,/0,/0,
	"SIOCATMARK"	/* 7 */
};

/* Socket types */
char sock_types[] = {
	/0,
	"SOCK_STREAM",     /* 1 */
	"SOCK_DGRAM",      /* 2 */
	"SOCK_RAW",        /* 3 */
	"SOCK_RDM",        /* 4 - reliably-delivered message */
	"SOCK_SEQPACKET"   /* 5 - sequenced packet stream */
};

/*
 * Option flags per-socket.
 *
 *  Two not here:
 *    PVD_CONFIG  0x3001
 *    TCP_NODELAY 0x0001
 */
char sock_opts1[] = {
	"SO_DEBUG",			// 0x0001 - turn on debugging info recording 
	"SO_ACCEPTCONN",    // 0x0002 - socket has had listen() 
	"SO_REUSEADDR",     // 0x0004 - allow local address reuse 
	"SO_KEEPALIVE",     // 0x0008 - keep connections alive 
	"SO_DONTROUTE",     // 0x0010 - just use interface addresses 
	"SO_BROADCAST",		// 0x0020 - permit sending of broadcast msgs 
	"SO_USELOOPBACK",	// 0x0040 - bypass hardware when possible 
	"SO_LINGER",        // 0x0080 - linger on close if data present 
	"SO_OOBINLINE"		// 0x0100 - leave received OOB data in line 
};
#define MAX_OPTS1
char sock_opts2[] = {
	"SO_SNDBUF",       // 0x1001 - send buffer size 
	"SO_RCVBUF",       // 0x1002 - receive buffer size 
	"SO_SNDLOWAT",     // 0x1003 - send low-water mark 
	"SO_RCVLOWAT",     // 0x1004 - receive low-water mark 
	"SO_SNDTIMEO",     // 0x1005 - send timeout 
	"SO_RCVTIMEO",     // 0x1006 - receive timeout 
	"SO_ERROR",        // 0x1007 - get error status and clear 
	"SO_TYPE"          // 0x1008 - get socket type 
	"SO_DONTLINGER"		// 0xFF6F - (~SO_LINGER)
};

char sock_opts3[] = {
	"SO_GROUP_ID",       // 0x2001 - ID of a socket group
	"SO_GROUP_PRIORITY", // 0x2002 - the relative priority within a group
	"SO_MAX_MSG_SIZE",   // 0x2003 - maximum message size 
	"SO_PROTOCOL_INFOA", // 0x2004 - WSAPROTOCOL_INFOA structure 
	"SO_PROTOCOL_INFOW"  // 0x2005 - WSAPROTOCOL_INFOW structure 
};
char sock_nodelay = {"TCP_NODELAY"};
char sock_pvd_cfg = {"PVD_CONFIG");
char sock_optx[]  = {"Unknown Option");



/*
 * Address families.
 */
#define        0               /* unspecified */
/*
 * Although  AF_UNSPEC  is  defined for backwards compatibility, using
 * AF_UNSPEC for the "af" parameter when creating a socket is STRONGLY
 * DISCOURAGED.    The  interpretation  of  the  "protocol"  parameter
 * depends  on the actual address family chosen.  As environments grow
 * to  include  more  and  more  address families that use overlapping
 * protocol  values  there  is  more  and  more  chance of choosing an
 * undesired address family when AF_UNSPEC is used.
 */
char addr_fams[] = {
	"AF_UNSPEC",		//0 unspecified 
	"AF_UNIX",         // 1 local to host (pipes, portals) 
	"AF_INET",         // 2 internetwork: UDP, TCP, etc. 
	"AF_IMPLINK",      // 3 arpanet imp addresses 
	"AF_PUP",          // 4 pup protocols: e.g. BSP 
	"AF_CHAOS",        // 5 mit CHAOS protocols 
	"AF_IPX",          // 6 IPX protocols: IPX, SPX, etc. 
	"AF_OSI",          // 7 OSI is ISO 
	"AF_ECMA",         // 8 european computer manufacturers 
	"AF_DATAKIT",      // 9  datakit protocols 
	"AF_CCITT",        // 10  CCITT protocols, X.25 etc 
	"AF_SNA",          // 11  IBM SNA 
	"AF_DECnet",       // 12  DECnet 
	"AF_DLI",          // 13  Direct data link interface 
	"AF_LAT",          // 14  LAT 
	"AF_HYLINK",       // 15  NSC Hyperchannel 
	"AF_APPLETALK",    // 16  AppleTalk 
	"AF_NETBIOS",      // 17  NetBios-style addresses 
	"AF_VOICEVIEW",    // 18  VoiceView 
	"AF_FIREFOX",      // 19  Protocols from Firefox 
	"AF_UNKNOWN1",     // 20  Somebody is using this! 
	"AF_BAN",          // 21  Bnnyan 
	"AF_ATM",          // 22  Native ATM Services 
	"AF_INET6"         // 23  Internetwork Version 6 
}


char recv_flags [] = {
	/0,
	"MSG_OOB",         // 0x1 - process out-of-band data 
	"MSG_PEEK",        // 0x2 - peek at incoming message 
	"MSG_DONTROUTE",   // 0x4 - send without using routing tables
	/0,
	"MSG_INTERRUPT",   // 0x10   - send/recv in the interrupt context */
	"MSG_PARTIAL"      // 0x8000 - partial send or recv for message xport */
};

char fd_events [] = {
	/0,
	"FD_READ",		// 0x1 
	"FD_WRITE",		// 0x2
	"FD_OOB",		// 0x4
	"FD_ACCEPT",	// 0x8
	"FD_CONNECT",	// 0x10
	"FD_CLOSE",		// 0x20
	"FD_QOS",		// 0x40
	"FD_GROUP_QOS"	// 0x80
};


/*
 * WinSock 2 extension -- manifest constants for return values of the condition function
 */
char cond_ret[] = {
	"CF_ACCEPT", // 0x0000
	"CF_REJECT", // 0x0001
	"CF_DEFER"   // 0x0002
};

/*
 * WinSock 2 extension -- manifest constants for shutdown()
 */
 char how[] = {
	"SD_RECEIVE", // 0x00
	"SD_SEND",    // 0x01
	"SD_BOTH"     // 0x02
}

/*
 * WinSock 2 extension -- data type and manifest constants for socket groups
 */
char group_type[] = {
	"none",
	"SG_UNCONSTRAINED_GROUP",   // 0x01
	"SG_CONSTRAINED_GROUP",     // 0x02
};

char prot_type[] = {
	"LAYERED_PROTOCOL",   // 0
	"BASE_PROTOCOL",      // 1
};

/* Flag bit definitions for dwProviderFlags */
char prov_flags [] = {
	"PFL_MULTIPLE_PROTO_ENTRIES",          // 0x00000001
	"PFL_RECOMMENDED_PROTO_ENTRY",         // 0x00000002
	"PFL_HIDDEN",                          // 0x00000004
	"PFL_MATCHES_PROTOCOL_ZERO"            // 0x00000008
};

/* Flag bit definitions for dwServiceFlags1 */
char srvc_flags [] = {
	"XP1_CONNECTIONLESS",                  // 0x00000001
	"XP1_GUARANTEED_DELIVERY",             // 0x00000002
	"XP1_GUARANTEED_ORDER",                // 0x00000004
	"XP1_MESSAGE_ORIENTED",                // 0x00000008
	"XP1_PSEUDO_STREAM",                   // 0x00000010
	"XP1_GRACEFUL_CLOSE",                  // 0x00000020
	"XP1_EXPEDITED_DATA",                  // 0x00000040
	"XP1_CONNECT_DATA",                    // 0x00000080
	"XP1_DISCONNECT_DATA",                 // 0x00000100
	"XP1_SUPPORT_BROADCAST",               // 0x00000200
	"XP1_SUPPORT_MULTIPOINT",              // 0x00000400
	"XP1_MULTIPOINT_CONTROL_PLANE",        // 0x00000800
	"XP1_MULTIPOINT_DATA_PLANE",           // 0x00001000
	"XP1_QOS_SUPPORTED",                   // 0x00002000
	"XP1_INTERRUPT",                       // 0x00004000
	"XP1_UNI_SEND",                        // 0x00008000
	"XP1_UNI_RECV",                        // 0x00010000
	"XP1_IFS_HANDLES",                     // 0x00020000
	"XP1_PARTIAL_MESSAGE"                  // 0x00040000
};

char byte_order[] = {
	"BIGENDIAN",                           // 0x0000
	"LITTLEENDIAN",                        // 0x0001
};

char sec_prot[] = {
	"SECURITY_PROTOCOL_NONE"               // 0x0000
};

/*
 * WinSock 2 extension -- manifest constants for WSAJoinLeaf()
 */
char join_cmd [] = {
	"JL_SENDER_ONLY",    // 0x01
	"JL_RECEIVER_ONLY",  // 0x02
	"JL_BOTH",           // 0x04
};

/*
 * WinSock 2 extension -- manifest constants for WSASocket()
 */
char wsock_flags [] = {
	"WSA_FLAG_OVERLAPPED",           // 0x01
	"WSA_FLAG_MULTIPOINT_C_ROOT",    // 0x02
	"WSA_FLAG_MULTIPOINT_C_LEAF",    // 0x04
	"WSA_FLAG_MULTIPOINT_D_ROOT",    // 0x08
	"WSA_FLAG_MULTIPOINT_D_LEAF"     // 0x10
};

/*
 * WinSock 2 extension -- manifest constants for WSAIoctl()
 */
char wioctl_type[] = {
	"IOC_UNIX",                      // 0x00000000
	"IOC_WS2",                       // 0x08000000
	"IOC_PROTOCOL",                  // 0x10000000
	"IOC_VENDOR"                     // 0x18000000
};

char wioctl_cmds[] = {
	/0,
	"SIO_ASSOCIATE_HANDLE",          
	"SIO_ENABLE_CIRCULAR_QUEUEING",
	"SIO_FIND_ROUTE",
	"SIO_FLUSH",              
	"SIO_GET_BROADCAST_ADDRESS",
	"SIO_GET_EXTENSION_FUNCTION_POINTER",
	"SIO_GET_QOS",
	"SIO_GET_GROUP_QOS",
	"SIO_MULTIPOINT_LOOPBACK",
	"SIO_MULTICAST_SCOPE",
	"SIO_SET_QOS",  
	"SIO_SET_GROUP_QOS",
	"SIO_TRANSLATE_HANDLE"
};

/*
 * WinSock 2 extension -- manifest constants for SIO_TRANSLATE_HANDLE ioctl
 */
char handle_type [] = {
	"TH_NETDEV",     //   0x00000001
	"TH_TAPI"        // 0x00000002
};


/*
 * Service Install Flags
 */
char servinst [] = {
	/0,
	"SERVICE_MULTIPLE"       // (0x00000001)
};


/*
 * Name Spaces
 */
char name_spaces [] = {
	"NS_ALL",                      // (0)
	"NS_SAP",                      // (1)
	"NS_NDS",                      // (2)
	"NS_PEER_BROWSE",              // (3)
	/0, /0, /0, /0, /0, /0,
	"NS_TCPIP_LOCAL",              // (10)
	"NS_TCPIP_HOSTS",              // (11)
	"NS_DNS",                      // (12)
	"NS_NETBT",                    // (13)
	"NS_WINS",                     // (14)
	/0, /0, /0, /0, /0,
	"NS_NBP",                      // (20)
	/0, /0, /0, /0, /0, /0, /0, /0, /0
	"NS_MS",                       // (30)
	"NS_STDA",                     // (31)
	"NS_NTDS",                     // (32)
	/0, /0, /0, /0, /0, /0, /0,
	"NS_X500",                     // (40)
	"NS_NIS",                      // (41)
	"NS_NISPLUS",                  // (42)
	/0, /0, /0, /0, /0, /0, /0,
	"NS_WRQ"                       // (50)
};

/*
 * Resolution flags for WSAGetAddressByName().
 * Note these are also used by the 1.1 API GetAddressByName, so
 * leave them around.
 */
char res_flags [] = {
	"RES_UNUSED_1",                // (0x00000001)
	"RES_FLUSH_CACHE",             // (0x00000002)
	"RES_SERVICE"	               // (0x00000004)
};


char nres_flags [] = {
	/0,
	"LUP_DEEP",                // 0x0001
	"LUP_CONTAINERS",          // 0x0002
	"LUP_NOCONTAINERS",        // 0x0004
	"LUP_NEAREST",             // 0x0008
	"LUP_RETURN_NAME",         // 0x0010
	"LUP_RETURN_TYPE",         // 0x0020
	"LUP_RETURN_VERSION",      // 0x0040
	"LUP_RETURN_COMMENT",      // 0x0080
	"LUP_RETURN_ADDR",         // 0x0100
	"LUP_RETURN_BLOB",         // 0x0200
	"LUP_RETURN_ALIASES",      // 0x0400
	"LUP_RETURN_QUERY_STRING", // 0x0800
	"LUP_FLUSHCACHE",		   // 0x1000
	"LUP_FLUSHPREVIOUS",       // 0x2000
	/0,
	"LUP_RES_SERVICE",         // 0x8000
	"LUP_RETURN_ALL"           // 0x0FF0
};



//
// Return flags
//
char result [] = {
	/0,
	"RESULT_IS_ALIAS"      // 0x0001
};


