/*++
  
  Copyright (c) 1996 Microsoft Corporation.

  Copyright (c) 1995 Intel Corp

  Copyright (c) 1996 Bob Quinn
  
  Module Name:
  
    nsp_func.cpp
  
  Abstract:
  
    Contains handler functions for all the NSP (namespace
	provider) function debug trace DLL (dt_dll.dll) hooks.
  
--*/


#include "nowarn.h"  /* turn off benign warnings */
#ifndef _WINSOCKAPI_
#define _WINSOCKAPI_   /* Prevent inclusion of winsock.h in windows.h */
#endif
#include <windows.h>
#include "nowarn.h"  /* some warnings may have been turned back on */
#include <ws2spi.h>

#include "handlers.h"
#include "dt_dll.h"
#include "dt.h"
/*++
  
  Note:

      This comment applies to all DTHandler_xxxx functions.
  
  Function Description:
  
      For the present, each handler function simply adds "xxxx
      called." to the end of the string passed in, and calls
      DTTextOut to output the whole string to the Debug Window.  This
      allows handler functions to be disabled easily, or to output
      more information (like the arguments) via multiple calls to
      DTTextOut, etc.  See the Debug/Trace documentation for more
      information. 
  
  Arguments:
  
      vl -- Variable used to strip arguments off the stack of 
      WSA[Pre|Post]ApiNotify, the calling function.

      ReturnValue -- A void pointer to the return value for the
      original API/SPI function.  Use a local variable to cast it to
      the appropriate type.

      LibraryName -- For API functions, this should just be
      "WinSock2", but for SPI functions it is the name of the service
      provider.  This is to distinguish among the multiple service
      providers that WinSock2 may be calling.

      Buffer -- A string containg some info about the thread and
      function call # of this call.  

      Index -- How many characters are in the above.
      
      BufLen -- How big is the buffer.

      PreOrPost -- TRUE if called by a WSAPreApiNotify, FALSE if
      called by WSAPostApiNotify.  
        
  Return Value:
  
      If the PreOrPost argument is FALSE, then the return value is
      ignored.  If it's TRUE, then the return value indicates to
      WSAPreApiNotify whether or not we want to short-circuit the API
      or SPI function.  TRUE means short-circuit it, FALSE mean don't.
  
--*/



BOOL CALLBACK
DTHandler_NSPLookupServiceBegin(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPGUID *lpProviderId = va_arg(vl, LPGUID *);
    LPWSAQUERYSETW *lpqsRestrictions = va_arg(vl, LPWSAQUERYSETW*);
    LPWSASERVICECLASSINFOW *lpServiceClassInfo =
        va_arg(vl, LPWSASERVICECLASSINFOW *);
    DWORD *dwControlFlags = va_arg(vl, DWORD*);
    LPHANDLE *lphLookup  = va_arg(vl, LPHANDLE*);
    
	int off = Index;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> NSPLookupServiceBegin()\r\n\t");
	} else {
		off += wsprintf(Buffer + off,
			"<- NSPLookupServiceBegin() returned: %d\r\n\t", *RetVal);
	}

	if (*lpProviderId) {
		if (!IsBadReadPtr (*lpProviderId, sizeof(GUID))) {
			off += wsprintf (Buffer + off, "lpProviderId: \r\n\t");
			off = FormatGUID (*lpProviderId, Buffer, off, "\t\t");
			off += wsprintf (Buffer + off, "\t");
		} else {
			off += wsprintf (Buffer + off, 
				"lpProviderId POINTER IS INVALID\r\n\t");
		}
	} else {
		off += wsprintf (Buffer + off, "lpProviderId is NULL\r\n\t");
	}

	if (*lpqsRestrictions) {
		off += wsprintf (Buffer + off, "lpqsRestrictions: \r\n\t");
		off = FormatQuerySetW (*lpqsRestrictions, Buffer, off, "\t\t");
		off += wsprintf (Buffer + off, "\t");
	} else {
		off += wsprintf (Buffer + off, "lpqsRestrictions is NULL\r\n\t");
	}
	if (*lpServiceClassInfo) {
		off += wsprintf (Buffer + off, "lpServiceClassInfo: \r\n\t");
		off = FormatServiceClassInfoW (*lpServiceClassInfo, Buffer, off, "\t\t");
		off += wsprintf (Buffer + off, "\t");
	} else {
		off += wsprintf (Buffer + off, "lpServiceClassInfo is NULL\r\n\t");
	}
	off += wsprintf (Buffer + off, "dwControlFlags: %d (0x%lx)\r\n\t",
		*dwControlFlags, *dwControlFlags);
	if (*lphLookup) {
		off += wsprintf (Buffer + off, "lphLookup: %d (0x%lx)\r\n",
			**lphLookup, **lphLookup);
	} else {
		off += wsprintf (Buffer + off, "lphLookup is NULL\r\n");
	}

    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}



BOOL CALLBACK
DTHandler_NSPLookupServiceNext(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    HANDLE *hLookup = va_arg(vl, HANDLE*);
    DWORD *dwControlFlags = va_arg(vl, DWORD*);
    LPDWORD *lpdwBufferLength = va_arg(vl, LPDWORD*);
    LPWSAQUERYSETW *lpqsResults = va_arg(vl, LPWSAQUERYSETW*);
 
	int off = Index;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> NSPLookupServiceNext()\r\n\t");
	} else {
		off += wsprintf(Buffer + off,
			"<- NSPLookupServiceNext() returned: %d\r\n\t", *RetVal);
	}
	off += wsprintf (Buffer + off, "hLookup: %d (0x%lx)\r\n\t",
		*hLookup, *hLookup);
	off += wsprintf (Buffer + off, "dwControlFlags: %d (0x%lx)\r\n\t",
		*dwControlFlags, *dwControlFlags);
	if (*lpdwBufferLength) {
		off += wsprintf (Buffer + off, "lpdwBufferLength: %d (0x%lx)\r\n\t",
			**lpdwBufferLength, **lpdwBufferLength);
	} else {
		off += wsprintf (Buffer + off, "lpdwBufferLength is NULL\r\n\t");
	}
 	if (*lpqsResults) {
		off += wsprintf (Buffer + off, "lpqsResults: \r\n\t");
		off = FormatQuerySetW (*lpqsResults, Buffer, off, "\t\t");
		off += wsprintf (Buffer + off, "\t");
	} else {
		off += wsprintf (Buffer + off, "lpqsResults is NULL\r\n");
	}
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}




BOOL CALLBACK
DTHandler_NSPLookupServiceEnd(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    HANDLE *hLookup = va_arg(vl, HANDLE*);
    
	int off = Index;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> NSPLookupServiceEnd()\r\n\t");
	} else {
		off += wsprintf(Buffer + off,
			"<- NSPLookupServiceEnd() returned: %d\r\n\t", *RetVal);
	}
	off += wsprintf (Buffer + off, "hLookup: %d (0x%lx)\r\n",
		*hLookup, *hLookup);

    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}



BOOL CALLBACK
DTHandler_NSPInstallServiceClass(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPGUID * lpProviderId = va_arg(vl, LPGUID *);
    LPWSASERVICECLASSINFOW *lpServiceClassInfo =
        va_arg(vl, LPWSASERVICECLASSINFOW*);
 
    wsprintf(Buffer + Index, "NSPInstallServiceClass() %s.\r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}



BOOL CALLBACK
DTHandler_NSPSetService(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPGUID * lpProviderId = va_arg(vl, LPGUID *);
    LPWSASERVICECLASSINFOW * lpServiceClassInfo =
        va_arg(vl, LPWSASERVICECLASSINFOW *);
    LPWSAQUERYSETW *lpqsRegInfo = va_arg(vl, LPWSAQUERYSETW*);
    WSAESETSERVICEOP *essOperation = va_arg(vl, WSAESETSERVICEOP*);
    DWORD *dwControlFlags = va_arg(vl, DWORD*);
 
    wsprintf(Buffer + Index, "NSPSetService() %s.\r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}



BOOL CALLBACK
DTHandler_NSPRemoveServiceClass(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPGUID *lpProviderId = va_arg(vl, LPGUID *);
    LPGUID *lpServiceClassId = va_arg(vl, LPGUID*);

    wsprintf(Buffer + Index, "NSPRemoveServiceClass() %s.\r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}



BOOL CALLBACK
DTHandler_NSPGetServiceClassInfo(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPGUID *lpProviderId = va_arg(vl, LPGUID*);
    LPDWORD *lpdwBufSize = va_arg(vl, LPDWORD*);
    LPWSASERVICECLASSINFOW *lpServiceClassInfo =
        va_arg(vl, LPWSASERVICECLASSINFOW*);

    wsprintf(Buffer + Index, "NSPGetServiceClassInfo() %s.\r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_NSPCleanup(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPGUID * lpProviderId = va_arg(vl, LPGUID *);

    wsprintf(Buffer + Index, "NSPCleanup() %s, \r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}

