/*++
  
  Copyright (c) 1996 Microsoft Corporation.

  Copyright (c) 1995 Intel Corp

  Copyright (c) 1996 Bob Quinn
  
  Module Name:
  
    wsa2func.cpp
  
  Abstract:
  
    Contains handler functions for all new WinSock 2
	function debug trace DLL (dt_dll.dll) hooks.
  
--*/

#include "nowarn.h"  /* turn off benign warnings */
#ifndef _WINSOCKAPI_
#define _WINSOCKAPI_   /* Prevent inclusion of winsock.h in windows.h */
#endif
#include <windows.h>
#include "nowarn.h"  /* some warnings may have been turned back on */
#include <ws2spi.h>

#include "handlers.h"
#include "dt_dll.h"
#include "dt.h"

extern char *fd_events [];

/* WSAIoctl constants */
#define WSA_IOC_BUF   0xE0000000
#define WSA_IOC_IN    0x80000000
#define WSA_IOC_OUT   0x40000000
#define WSA_IOC_BOTH  0xC0000000
#define WSA_IOC_VOID  0x20000000
#define WSA_IOC_TYPE  0x18000000
#define WSA_IOC_UNIX  0x00000000
#define WSA_IOC_WS2   0x08000000
#define WSA_IOC_AF    0x10000000
#define WSA_IOC_VNDR  0x18000000
#define WSA_IOC_ID    0x07FF0000
#define WSA_IOC_CODE  0x0000FFFF


/* BSD ioctl() string values (declared in BSD_FUNC.CPP) */
extern char *ioctl_ioc[];
extern char *ioctl_fio[];
extern char *ioctl_sio[];

/* WSAIoctl() string values */
char *wioctl_type[] = {
	"IOC_UNIX",                      // 0x00000000
	"IOC_WS2",                       // 0x08000000
	"IOC_PROTOCOL",                  // 0x10000000
	"IOC_VENDOR"                     // 0x18000000
};

char *wioctl_cmds[] = {
	"unknown",
	"SIO_ASSOCIATE_HANDLE",				  // 1 (code value)          
	"SIO_ENABLE_CIRCULAR_QUEUEING",		  // 2
	"SIO_FIND_ROUTE",	    			  // 3
	"SIO_FLUSH",						  // 4
	"SIO_GET_BROADCAST_ADDRESS",		  // 5
	"SIO_GET_EXTENSION_FUNCTION_POINTER", // 6
	"SIO_GET_QOS",						  // 7
	"SIO_GET_GROUP_QOS",				  // 8
	"SIO_MULTIPOINT_LOOPBACK",			  // 9
	"SIO_MULTICAST_SCOPE",				  // 10
	"SIO_SET_QOS",						  // 11
	"SIO_SET_GROUP_QOS",				  // 12
	"SIO_TRANSLATE_HANDLE"				  // 13
};

/*
 * Service Address Registration and Deregistration Data Types.
 */
char *setservops [] = {
    "RNRSERVICE_REGISTER",   // 0
    "RNRSERVICE_DEREGISTER", // 1
    "RNRSERVICE_DELETE"      // 2
};

 /*++
  
  Note:

      This comment applies to all DTHandler_xxxx functions.
  
  Function Description:
  
      For the present, each handler function simply adds "xxxx
      called." to the end of the string passed in, and calls
      DTTextOut to output the whole string to the Debug Window.  This
      allows handler functions to be disabled easily, or to output
      more information (like the arguments) via multiple calls to
      DTTextOut, etc.  See the Debug/Trace documentation for more
      information. 
  
  Arguments:
  
      vl -- Variable used to strip arguments off the stack of 
      WSA[Pre|Post]ApiNotify, the calling function.

      ReturnValue -- A void pointer to the return value for the
      original API/SPI function.  Use a local variable to cast it to
      the appropriate type.

      LibraryName -- For API functions, this should just be
      "WinSock2", but for SPI functions it is the name of the service
      provider.  This is to distinguish among the multiple service
      providers that WinSock2 may be calling.

      Buffer -- A string containg some info about the thread and
      function call # of this call.  

      Index -- How many characters are in the above.
      
      BufLen -- How big is the buffer.

      PreOrPost -- TRUE if called by a WSAPreApiNotify, FALSE if
      called by WSAPostApiNotify.  
        
  Return Value:
  
      If the PreOrPost argument is FALSE, then the return value is
      ignored.  If it's TRUE, then the return value indicates to
      WSAPreApiNotify whether or not we want to short-circuit the API
      or SPI function.  TRUE means short-circuit it, FALSE mean don't.
  
--*/

BOOL CALLBACK
DTHandler_WSAAccept(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    SOCKET *RetVal = (SOCKET *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    struct sockaddr FAR **addr = va_arg(vl, struct sockaddr FAR **);
    LPINT *addrlen = va_arg(vl, LPINT *);
    LPCONDITIONPROC *lpfnCondition = va_arg(vl, LPCONDITIONPROC *);
    DWORD *dwCallbackData = va_arg(vl, DWORD *);

    int off = Index;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, 
			"-> WSAAccept()\r\n\tsocket = %d\r\n\t",*s);
		off += wsprintf (Buffer + off, "addrlen: %d\r\n\t", *addrlen);
		if (*lpfnCondition) {
			off += wsprintf (Buffer + off, "lpfnCondition: 0x%08x\r\n\t");
		} else {
			off += wsprintf(Buffer + off, "lpfnCondition is NULL\r\n\t");
		}
		wsprintf (Buffer + off, "dwCallbackData: %d (0x%x)\r\n");
	} else {
		off += wsprintf(Buffer + off, "<- WSAAccept() returned: %d\r\n",
            *RetVal);
		if (*addr) {
			off += FormatSockaddrIn((LPSOCKADDR_IN)*addr, Buffer, off);
			off += wsprintf (Buffer + off, "\t");
		} else {
			off += wsprintf(Buffer + off, "addr is NULL\r\n\t");
		}
		off += wsprintf (Buffer + off, "addrlen: %d\r\n\t", *addrlen);
	}
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}

BOOL CALLBACK
DTHandler_WSACloseEvent(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    BOOL *RetVal = (BOOL *)ReturnValue;
    WSAEVENT *hEvent = va_arg(vl, WSAEVENT *);

    if (PreOrPost) {
		wsprintf(Buffer + Index, 
			"-> WSACloseEvent()\r\nhEvent = %d\r\n",*hEvent);
	} else {
		wsprintf(Buffer + Index,
			"<- WSCloseEvent() returned: %s\r\n", 
			(*RetVal ? "TRUE" : "FALSE"));
	}
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAConnect(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    struct sockaddr FAR **name = va_arg(vl, struct sockaddr FAR **);
    int *namelen = va_arg(vl, int *);
    LPWSABUF *lpCallerData = va_arg(vl, LPWSABUF *);
    LPWSABUF *lpCalleeData = va_arg(vl, LPWSABUF *);
    LPQOS *lpSQOS = va_arg(vl, LPQOS *);
    LPQOS *lpGQOS = va_arg(vl, LPQOS *);
	int off;

    off = Index;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, 
			"-> WSAConnect()\r\n\tsocket = %d\r\n\t",*s);
		if (*name) {
			off += FormatSockaddrIn((LPSOCKADDR_IN)*name, Buffer, off);
			off += wsprintf (Buffer + off, "\t");
		} else {
			off += wsprintf(Buffer + off,"name is NULL\r\n");
		}
		off += wsprintf (Buffer + off, "namelen: %d\r\n\t", *namelen);
		if (lpCallerData) {
			DTTextOut (DebugWindow, LogFileHandle, Buffer, OutputStyle);
			off = Index;
			off += wsprintf (Buffer + off, "CallerData: \r\n\t\t");
//			off += FormatWSABuf(*lpCallerData, Buffer, off);
		} else {
			off += wsprintf(Buffer + off, "lpCallerData is NULL\r\n\t");
		}
	} else {
		off += wsprintf(Buffer + off, "<- WSAConnect() returned: %d\r\n",
            *RetVal);
		if (lpCalleeData) {
			DTTextOut (DebugWindow, LogFileHandle, Buffer, OutputStyle);
			off = Index;
			off += wsprintf (Buffer + off, "CalleeData: \r\n\t\t");
//			off += FormatWSABuf(*lpCalleeData, Buffer, off);
		} else {
			off += wsprintf(Buffer + off, "lpCalleeData is NULL\r\n\t");
		}
	}

	if (*lpSQOS) {
		off += FormatQOS (*lpSQOS, Buffer, off);
	} else {
		off += wsprintf (Buffer + off, "lpSQOS is NULL\r\n\t");
	}

	if (*lpGQOS) {
		off += FormatQOS (*lpGQOS, Buffer, off);
	} else {
		off += wsprintf (Buffer + off, "lpGQOS is NULL\r\n\t");
	}

    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSACreateEvent(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    WSAEVENT *RetVal = (WSAEVENT *)ReturnValue;
    
    if (PreOrPost) {
		wsprintf(Buffer + Index, 
			"-> WSACreateEvent()\r\n\r\n");
	} else {
		wsprintf(Buffer + Index,
			"<- WSCreateEvent() returned: %d\r\n", *RetVal);
	}
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSADuplicateSocketA(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    DWORD *dwProcessId = va_arg(vl, DWORD *);
    LPWSAPROTOCOL_INFOA *lpProtocolInfo = va_arg(vl, LPWSAPROTOCOL_INFOA *);

    int off = Index;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSADuplicateSocketA()\r\n\t");
	} else {
		off += wsprintf(Buffer + off, 
			"<- WSADuplicateSocketA() returned: %d\r\n\t",*RetVal);
	}
	off += wsprintf (Buffer + off, "socket: %d\r\n\t", *s);
	off += wsprintf (Buffer + off, "PID:    0x%08\r\n", *dwProcessId);
	if (!PreOrPost) {
		if (*lpProtocolInfo) {
			off += wsprintf(Buffer + off, "\tlpProtocolInfo:\r\n\t\t");
			FormatProtocolInfo(*lpProtocolInfo, Buffer, off);
		} else {
			off += wsprintf(Buffer + off, "\tlpProtocolInfo is NULL\r\n");
		}
	}

    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSADuplicateSocketW(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    DWORD *dwProcessId = va_arg(vl, DWORD *);
    LPWSAPROTOCOL_INFOW *lpProtocolInfo = va_arg(vl, LPWSAPROTOCOL_INFOW *);

    int off = Index;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSADuplicateSocketA()\r\n\t");
	} else {
		off += wsprintf(Buffer + off, 
			"<- WSADuplicateSocketA() returned: %d\r\n\t",*RetVal);
	}
	off += wsprintf (Buffer + off, "socket: %d\r\n\t", *s);
	off += wsprintf (Buffer + off, "PID:    0x%08\r\n", *dwProcessId);
	if (!PreOrPost) {
		if (*lpProtocolInfo) {
			off += wsprintf(Buffer + off, "\tlpProtocolInfo:\r\n\t\t");
//			FormatProtocolInfo(*lpProtocolInfo, Buffer, off);
		} else {
			off += wsprintf(Buffer + off, "\tlpProtocolInfo is NULL\r\n");
		}
	}

    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAEnumNetworkEvents(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    WSAEVENT *hEventObject = va_arg(vl, WSAEVENT *);
    LPWSANETWORKEVENTS *lpNetworkEvents = va_arg(vl, LPWSANETWORKEVENTS *);

	int off = Index;
    if (PreOrPost) {
		off += wsprintf(Buffer + off, "<- WSAEnumNetworkEvents()\r\n\t");
	} else {
		off += wsprintf(Buffer + off, 
			"-> WSAEnumNetworkEvents() returned: %d\r\n\t",*RetVal);
	}
	off += wsprintf(Buffer + off, "socket: %d\r\n\t", *s);
	off += wsprintf(Buffer + off, "hEventObject: %d\r\n\t", *hEventObject);
	if (*lpNetworkEvents) {
		off = FormatWSANetworkEvents(*lpNetworkEvents, Buffer, off);
		wsprintf (Buffer + off, "\r\n");
	} else {
		wsprintf (Buffer + off, "lpNetworkEvents is NULL\r\n");
	}
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAEnumProtocolsA(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPINT *lpiProtocols = va_arg(vl, LPINT *);
    LPWSAPROTOCOL_INFOA *lpProtocolBuffer = va_arg(vl, LPWSAPROTOCOL_INFOA *);
    LPDWORD *lpdwBufferLength = va_arg(vl, LPDWORD *);
	int off;

	off = Index;
	if (bOutputEnabled && (dDebugFlags & DT_SOCKET)) {
		int i,j=0;

		if (PreOrPost) {
			off += wsprintf(Buffer + off, 
				"-> WSAEnumProtocolsA()\r\n\t");
			if (*lpiProtocols)		// Number of WSAProtocol structs
				j = **lpiProtocols;
		} else {
			off += wsprintf(Buffer + off, 
				"<- WSAEnumProtocolsA() returned = %d\r\n\t",
				*RetVal);
			if (*RetVal)	  		// Number of WSAProtocol structs
				j = *RetVal;
		}

		/* Number of WSAProtocol structures in buffer */
		if (*lpiProtocols) {
			off += wsprintf(Buffer + off, 
				"lpiProtocols=%d\r\n\t",
				**lpiProtocols);
		} else {
			off += wsprintf(Buffer + off,
				"lpiProtocols=NULL\r\n\t");
		}

		/* Protocol Buffer */
		if (*lpProtocolBuffer) {
			DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
			off = Index;
			for (i=0; i<j; i++) {
				FormatProtocolInfo
					(((*lpProtocolBuffer)+i),Buffer, off);
				DTTextOut(DebugWindow, LogFileHandle, Buffer+off, OutputStyle);
			}
			off += wsprintf (Buffer + off, "\t");
		} else {
			off += wsprintf (Buffer + off, "lpProtocolBuffer=NULL\r\n\t");
		}

		/* Buffer length */
		if (*lpdwBufferLength) {
			off += wsprintf(Buffer + off, 
				"lpdwBufferLength=%lu\r\n",
				**lpdwBufferLength);
		} else {
			off += wsprintf(Buffer + off,
				"lpdwBufferLength=NULL\r\n");
		}
		DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
	}
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAEnumProtocolsW(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPINT *lpiProtocols = va_arg(vl, LPINT *);
    LPWSAPROTOCOL_INFOW *lpProtocolBuffer = va_arg(vl, LPWSAPROTOCOL_INFOW *);
    LPDWORD *lpdwBufferLength = va_arg(vl, LPDWORD *);

	int off;

	off = Index;
	if (bOutputEnabled && (dDebugFlags & DT_SOCKET)) {
		int i,j=0;

		if (PreOrPost) {
			off += wsprintf(Buffer + off, 
				"-> WSAEnumProtocolsA()\r\n\t");
			if (*lpiProtocols)		// Number of WSAProtocol structs
				j = **lpiProtocols;
		} else {
			off += wsprintf(Buffer + off, 
				"<- WSAEnumProtocolsA() returned = %d\r\n\t",
				*RetVal);
			if (*RetVal)	  		// Number of WSAProtocol structs
				j = *RetVal;
		}

		/* Number of WSAProtocol structures in buffer */
		if (*lpiProtocols) {
			off += wsprintf(Buffer + off, 
				"lpiProtocols=%d\r\n\t",
				**lpiProtocols);
		} else {
			off += wsprintf(Buffer + off,
				"lpiProtocols=NULL\r\n\t");
		}

		/* Protocol Buffer */
		if (*lpProtocolBuffer) {
			DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
			off = Index;
			for (i=0; i<j; i++) {
//				FormatProtocolInfo
//					(((*lpProtocolBuffer)+i),Buffer, off);
				DTTextOut(DebugWindow, LogFileHandle, Buffer+off, OutputStyle);
			}
			off += wsprintf (Buffer + off, "\t");
		} else {
			off += wsprintf (Buffer + off, "lpProtocolBuffer=NULL\r\n\t");
		}

		/* Buffer length */
		if (*lpdwBufferLength) {
			off += wsprintf(Buffer + off, 
				"lpdwBufferLength=%lu\r\n",
				**lpdwBufferLength);
		} else {
			off += wsprintf(Buffer + off,
				"lpdwBufferLength=NULL\r\n");
		}
		DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
	}
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAEventSelect(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    WSAEVENT *hEventObject = va_arg(vl, WSAEVENT *);
    long *lNetworkEvents = va_arg(vl, long *);

	int off = Index;

    if (PreOrPost) {
		int i;
		off += wsprintf(Buffer + off, "<- WSAEventSelect()\r\n\t");
		off += wsprintf(Buffer + off, "socket = %d\r\n\t", *s);
		off += wsprintf(Buffer + off, "hEventObject = %d\r\n\t", 
			*hEventObject);
		off += wsprintf(Buffer + off,"lNetworkEvents: ");
		if (*lNetworkEvents) {
			for (i=0; i<FD_MAX_EVENTS; i++) {
				if (*lNetworkEvents & (1<<i))
					off += wsprintf(Buffer + off, "%s ", fd_events[i]); 
			}
		} else {
			off += wsprintf (Buffer + off, "<none>");
		}
		wsprintf(Buffer + off, "\r\n");
	} else {
		wsprintf(Buffer + off, "-> WSAEventSelect() returned: %d\r\n", *RetVal);
	}

    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}

BOOL CALLBACK
DTHandler_WSAGetOverlappedResult(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    BOOL *RetVal = (BOOL *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    LPWSAOVERLAPPED *lpOverlapped = va_arg(vl, LPWSAOVERLAPPED *);
    LPDWORD *lpcbTransfer = va_arg(vl, LPDWORD *);
    BOOL *fWait = va_arg(vl, BOOL *);
    LPDWORD *lpdwFlags = va_arg(vl, LPDWORD *);

	int off = Index;
    if (PreOrPost) {
		off += wsprintf(Buffer + off, "<- WSAGetOverlappedResult()\r\n\t");
	} else {
		off += wsprintf(Buffer + off, 
			"-> WSAGetOverlappedResult() returned: %s\r\n\t",
			*RetVal ? "TRUE" : "FALSE");
	}
	off += wsprintf(Buffer + off, "socket: %d\r\n\t", *s);
	if (*lpOverlapped) {
		off = FormatWSAOverlapped (*lpOverlapped, Buffer, off);
		off += wsprintf (Buffer + off, "\t");
	} else {
		off += wsprintf (Buffer + off, "lpOverlapped is NULL\r\n\t");
	}
	off += wsprintf ("fWait: %s\r\n\t", *fWait ? "TRUE" : "FALSE");
	if (*lpdwFlags) {
		wsprintf (Buffer + off, "lpdwFlags: %d (0x%lx)\r\n",
			*lpdwFlags, *lpdwFlags);
	} else {
		wsprintf (Buffer + off, "lpdwFlags is NULL\r\n");
	}

    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAGetQOSByName(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    BOOL *RetVal = (BOOL *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    LPWSABUF *lpQOSName = va_arg(vl, LPWSABUF *);
    LPQOS *lpQOS = va_arg(vl, LPQOS *);

	int off = Index;
    if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSAGetQOSByName()\r\n\t");
	} else {
		off += wsprintf(Buffer + off, 
			"<- WSAGetQOSByName() returned: %d\r\n\t",*RetVal);
	}
	off += wsprintf(Buffer + off, "socket: %d\r\n\t", *s);
	if (PreOrPost) {
		if (*lpQOSName) {
			off = FormatWSABuf (*lpQOSName, Buffer, off, PreOrPost);
			off += wsprintf (Buffer + off, "\t");
		} else {
			off += wsprintf (Buffer + off, "lpQOSName is NULL\r\n\t");
		}
	} else {
		if (*lpQOS) {
			off = FormatQOS (*lpQOS, Buffer, off);
			off += wsprintf (Buffer + off, "\t");
		} else {
			off += wsprintf (Buffer + off, "lpQOS is NULL\r\n\t");
		}
	}
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAHtonl(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    u_long *hostlong = va_arg(vl, u_long *);
    u_long FAR **lpnetlong = va_arg(vl, u_long FAR **);

	int off = Index;
    if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSAHtonl()\r\n\t");
	} else {
		off += wsprintf(Buffer + off, 
			"<- WSAHtonl() returned: %d\r\n\t",*RetVal);
	}
	off += wsprintf(Buffer + off, "socket: %d (0x%lx)\r\n\t", *s);
	if (*lpnetlong) {
		off += wsprintf (Buffer + off, "*lpnetlong: %d (0x%lx)\r\n", 
			**lpnetlong, **lpnetlong);
	} else {
		off += wsprintf (Buffer + off, "lpnetlong is NULL\r\n");
	}
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAHtons(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    u_short *hostshort = va_arg(vl, u_short *);
    u_short FAR **lpnetshort = va_arg(vl, u_short FAR **);

	int off = Index;
    if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSANtohl()\r\n\t");
	} else {
		off += wsprintf(Buffer + off, 
			"<- WSANtohl() returned: %d\r\n\t",*RetVal);
	}
	off += wsprintf(Buffer + off, "socket: %d\r\n\t", *s);

	off += wsprintf(Buffer + off, "hostshort: %d (0x%x)\r\n\t",
		*hostshort, *hostshort);
	if (*lpnetshort) {
		off += wsprintf(Buffer + off, "lpnetshort: %d (0x%x)\r\n",
			**lpnetshort, **lpnetshort);
	} else {
		off += wsprintf(Buffer + off, "lpnetshort is NULL\r\n");
	}

    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAIoctl(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    DWORD *dwIoControlCode = va_arg(vl, DWORD *);
    LPVOID *lpvInBuffer = va_arg(vl, LPVOID *);
    DWORD *cbInBuffer = va_arg(vl, DWORD *);
    LPVOID *lpvOutBuffer = va_arg(vl, LPVOID *);
    DWORD *cbOutBuffer = va_arg(vl, DWORD *);
    LPDWORD *lpcbBytesReturned = va_arg(vl, LPDWORD *);
    LPWSAOVERLAPPED *lpOverlapped = va_arg(vl, LPWSAOVERLAPPED *);
    LPWSAOVERLAPPED_COMPLETION_ROUTINE *lpCompletionRoutine = 
      va_arg(vl, LPWSAOVERLAPPED_COMPLETION_ROUTINE *);

    int off = Index, i;
	DWORD dwTmp;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSAIoctl()\r\n\t");
	} else {
		off += wsprintf(Buffer + off, "<- WSAIoctl() returned: %d\r\n\t",
            *RetVal);
	}

	off += wsprintf(Buffer + off, "socket: %d\r\n\t",*s);
	off += wsprintf(Buffer + off, "dwIoControlCode:\r\n\t\t");
	switch (*dwIoControlCode & WSA_IOC_BUF) {
		case WSA_IOC_VOID:
			i = 0;
			break;
		case WSA_IOC_OUT:
			i = 1;
			break;
		case WSA_IOC_IN:
			i = 2;
			break;
		case WSA_IOC_BOTH:
			i = 3;
			break;
		default:
			i = 0;
	}
	off += wsprintf (Buffer + off, "Buffer type:  %s\r\n\t\t", ioctl_ioc[i]);

	dwTmp = *dwIoControlCode & WSA_IOC_CODE;
	switch (*dwIoControlCode & WSA_IOC_TYPE) {
		case WSA_IOC_UNIX:
			i = 0;
			dwTmp &= 0x0000000F;
			if (((dwTmp >  0)   && (dwTmp <= 7)) ||
				((dwTmp >= 0xd) && (dwTmp <= 0xf))) {
				off += wsprintf (Buffer + off, 
					"Command code: %s (0x%lx)\r\n\t\t", 
					(dwTmp <= 7) ? ioctl_sio[dwTmp] : ioctl_fio[dwTmp-13], 
					dwTmp);
			} else {
				off += wsprintf (Buffer + off,
					"Command code: unknown value %d (0x%lx)\r\n\t\t",
					dwTmp, dwTmp);
			}
			break;
		case WSA_IOC_WS2:
			i = 1;
			if ((dwTmp >  0) && (dwTmp <= 13)) {
				off += wsprintf (Buffer + off, 
					"Command code: %s (0x%lx)\r\n\t\t", 
					wioctl_cmds[dwTmp], dwTmp);
			} else {
				off += wsprintf (Buffer + off,
					"Command code: unknown value %d (0x%lx)\r\n\t\t",
					dwTmp, dwTmp);
			}
			break;
		case WSA_IOC_AF:
			i = 2;
			off += wsprintf (Buffer + off,
				"Command code: vendor/af value %d (0x%lx)\r\n\t\t",
				dwTmp, dwTmp);
			break;
		default:
			i = 0;
			off += wsprintf (Buffer + off,
				"Command code: unknown value %d (0x%lx)\r\n\t\t",
				dwTmp, dwTmp);
	}
	off += wsprintf (Buffer + off, "Command type: %s\r\n\t\t", wioctl_type[i]);

	dwTmp = *dwIoControlCode & WSA_IOC_ID;
	off += wsprintf (Buffer + off, "Vendor/AF ID: %d (0x%lx)\r\n\t", 
		dwTmp, dwTmp);

	off += wsprintf (Buffer + off, "lpvInBuffer: %d (0x%lx)\r\n\t",
		*lpvInBuffer, *lpvInBuffer);

	off += wsprintf (Buffer + off, "cbInBuffer: %d (0x%lx)\r\n\t",
		*cbInBuffer, *cbInBuffer);

	off += wsprintf (Buffer + off, "lpvOutBuffer: %d (0x%lx)\r\n\t",
		*lpvOutBuffer, *lpvOutBuffer);

	off += wsprintf (Buffer + off, "cbOutBuffer: %d (0x%lx)\r\n\t",
		*cbOutBuffer, *cbInBuffer);

	if (*lpcbBytesReturned) {
		off += wsprintf (Buffer + off, "lpcbBytesReturned: %d (0x%lx)\r\n\t",
			**lpcbBytesReturned, **lpcbBytesReturned);
	} else {
		off += wsprintf (Buffer + off, "lpcbBytesReturned is NULL\r\n\t");
	}
	if (*lpOverlapped) {
		off += wsprintf (Buffer + off, "lpOverlapped.hEvent: %d (0x%lx)\r\n\t",
			(*lpOverlapped)->hEvent, (*lpOverlapped)->hEvent);
	} else {
		off += wsprintf (Buffer + off, "lpOverlapped is NULL\r\n\t");
	}
	if (*lpCompletionRoutine) {
		wsprintf (Buffer + off, "lpCompletionRoutine: 0x%lx\r\n",
			lpCompletionRoutine);
	} else {
		wsprintf (Buffer + off, "lpCompletionRoutine is NULL\r\n");
	}
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}

BOOL CALLBACK
DTHandler_WSAJoinLeaf(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    SOCKET *RetVal = (SOCKET *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    struct sockaddr FAR **name = va_arg(vl, struct sockaddr FAR **);
    int *namelen = va_arg(vl, int *);
    LPWSABUF *lpCallerData = va_arg(vl, LPWSABUF *);
    LPWSABUF *lpCalleeData = va_arg(vl, LPWSABUF *);
    LPQOS *lpSQOS = va_arg(vl, LPQOS *);
    LPQOS *lpGQOS = va_arg(vl, LPQOS *);
    DWORD *dwFlags = va_arg(vl, DWORD *);
	
    int off = Index;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, 
			"-> WSAJoinLeaf()\r\n\tsocket = %d\r\n\t",*s);
		if (*name) {
			off += FormatSockaddrIn((LPSOCKADDR_IN)*name, Buffer, off);
			off += wsprintf (Buffer + off, "\t");
		} else {
			off += wsprintf(Buffer + off,"name is NULL\r\n");
		}
		off += wsprintf (Buffer + off, "namelen: %d\r\n\t", *namelen);
		if (lpCallerData) {
			DTTextOut (DebugWindow, LogFileHandle, Buffer, OutputStyle);
			off = Index;
			off += wsprintf (Buffer + off, "CallerData: \r\n\t\t");
//			off += FormatWSABuf(*lpCallerData, Buffer, off);
		} else {
			off += wsprintf(Buffer + off, "lpCallerData is NULL\r\n\t");
		}
	} else {
		off += wsprintf(Buffer + off, "<- WSAJoinLeaf() returned: %d\r\n",
            *RetVal);
		if (lpCalleeData) {
			DTTextOut (DebugWindow, LogFileHandle, Buffer, OutputStyle);
			off = Index;
			off += wsprintf (Buffer + off, "CalleeData: \r\n\t\t");
//			off += FormatWSABuf(*lpCalleeData, Buffer, off);
		} else {
			off += wsprintf(Buffer + off, "lpCalleeData is NULL\r\n\t");
		}
	}

	if (*lpSQOS) {
		off += FormatQOS (*lpSQOS, Buffer, off);
	} else {
		off += wsprintf (Buffer + off, "lpSQOS is NULL\r\n\t");
	}

	if (*lpGQOS) {
		off += FormatQOS (*lpGQOS, Buffer, off);
	} else {
		off += wsprintf (Buffer + off, "lpGQOS is NULL\r\n\t");
	}

	off += wsprintf (Buffer + off, "dwFlags: %d (0x%lx)\r\n",
		*dwFlags, *dwFlags);

    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSANtohl(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    u_long *netlong = va_arg(vl, u_long *);
    u_long FAR **lphostlong = va_arg(vl, u_long FAR **);

	int off = Index;
    if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSANtohl()\r\n\t");
	} else {
		off += wsprintf(Buffer + off, 
			"<- WSANtohl() returned: %d\r\n\t",*RetVal);
	}
	off += wsprintf(Buffer + off, "socket: %d\r\n\t", *s);

	off += wsprintf(Buffer + off, "netlong: %d (0x%lx)\r\n\t",
		*netlong, *netlong);
	if (*lphostlong) {
		off += wsprintf(Buffer + off, "lphostlong: %d (0x%lx)\r\n",
			**lphostlong, **lphostlong);
	} else {
		off += wsprintf(Buffer + off, "lphostlong is NULL\r\n");
	}

    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSANtohs(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    u_short *netshort = va_arg(vl, u_short *);
    u_short FAR **lphostshort = va_arg(vl, u_short FAR **);

	int off = Index;
    if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSANtohs()\r\n\t");
	} else {
		off += wsprintf(Buffer + off, 
			"<- WSANtohl() returned: %d\r\n\t",*RetVal);
	}
	off += wsprintf(Buffer + off, "socket: %d\r\n\t", *s);

	off += wsprintf(Buffer + off, "netshort: %d (0x%x)\r\n\t",
		*netshort, *netshort);
	if (*lphostshort) {
		off += wsprintf(Buffer + off, "lphostshort: %d (0x%x)\r\n",
			**lphostshort, **lphostshort);
	} else {
		off += wsprintf(Buffer + off, "lphostshort is NULL\r\n");
	}
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSARecv(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    LPWSABUF *lpBuffers = va_arg(vl, LPWSABUF *);
    DWORD *dwBufferCount = va_arg(vl, DWORD *);
    LPDWORD *lpNumberOfBytesRecvd = va_arg(vl, LPDWORD *);
    LPDWORD *lpFlags = va_arg(vl, LPDWORD *);
    LPWSAOVERLAPPED *lpOverlapped = va_arg(vl, LPWSAOVERLAPPED *);
    LPWSAOVERLAPPED_COMPLETION_ROUTINE *lpCompletionRoutine = 
      va_arg(vl, LPWSAOVERLAPPED_COMPLETION_ROUTINE *);

	int off = Index;
    if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSARecv()\r\n\t");
	} else {
		off += wsprintf(Buffer + off, 
			"<- WSARecv() returned: %d\r\n\t",*RetVal);
	}
	off += wsprintf(Buffer + off, "socket: %d\r\n\t", *s);
	if (*lpBuffers) {
		off = FormatWSABuf (*lpBuffers, Buffer, off, PreOrPost);
		off += wsprintf (Buffer + off, "\t");
	} else {
		off += wsprintf (Buffer + off, "lpBuffers is NULL\r\n\t");
	}
	if (*lpNumberOfBytesRecvd) {
		off += wsprintf (Buffer + off, "lpNumberOfBytesRecvd: %d (0x%lx)\r\n\t",
			**lpNumberOfBytesRecvd, **lpNumberOfBytesRecvd);
	} else {
		off += wsprintf (Buffer + off, "lpNumberOfBytesRecvd is NULL\r\n\t");
	}
	if (*lpFlags) {
		off += wsprintf (Buffer + off, "lpFlags: %d (0x%lx)\r\n\t",
			**lpFlags, **lpFlags);
	} else {
		off += wsprintf (Buffer + off, "lpFlags is NULL\r\n\t");
	}
	if (*lpOverlapped) {
		off += wsprintf (Buffer + off, "lpOverlapped.hEvent: %d (0x%lx)\r\n\t",
			(*lpOverlapped)->hEvent, (*lpOverlapped)->hEvent);
	} else {
		off += wsprintf (Buffer + off, "lpOverlapped is NULL\r\n\t");
	}
	if (*lpCompletionRoutine) {
		wsprintf (Buffer + off, "lpCompletionRoutine: 0x%lx\r\n",
			lpCompletionRoutine);
	} else {
		wsprintf (Buffer + off, "lpCompletionRoutine is NULL\r\n");
	}

    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSARecvDisconnect(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    LPWSABUF *lpInboundDisconnectData = va_arg(vl, LPWSABUF *);

	if (PreOrPost) {
		Index += wsprintf(Buffer + Index, "-> WSARecvDisconnect()\r\n\t");
		wsprintf (Buffer + Index, "socket: %d\r\n", *s);
	} else {
		Index += wsprintf(Buffer + Index, 
			"<- WSARecvDisconnect() returned: %d\r\n\t",*RetVal);
		Index = FormatWSABuf (*lpInboundDisconnectData, Buffer, Index, TRUE);
		wsprintf (Buffer + Index, "\r\n");
	}
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSARecvFrom(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    LPWSABUF *lpBuffers = va_arg(vl, LPWSABUF *);
    DWORD *dwBufferCount = va_arg(vl, DWORD *);
    LPDWORD *lpNumberOfBytesRecvd = va_arg(vl, LPDWORD *);
    LPDWORD *lpFlags = va_arg(vl, LPDWORD *);
    struct sockaddr FAR **lpFrom = va_arg(vl, struct sockaddr FAR **);
    LPINT *lpFromlen = va_arg(vl, LPINT *);
    LPWSAOVERLAPPED *lpOverlapped = va_arg(vl, LPWSAOVERLAPPED *);
    LPWSAOVERLAPPED_COMPLETION_ROUTINE *lpCompletionRoutine = 
      va_arg(vl, LPWSAOVERLAPPED_COMPLETION_ROUTINE *);

	int off = Index;
    if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSARecv()\r\n\t");
	} else {
		off += wsprintf(Buffer + off, 
			"<- WSARecv() returned: %d\r\n\t",*RetVal);
	}
	off += wsprintf(Buffer + off, "socket: %d\r\n\t", *s);
	if (*lpBuffers) {
		off = FormatWSABuf (*lpBuffers, Buffer, off, PreOrPost);
		off += wsprintf (Buffer + off, "\t");
	} else {
		off += wsprintf (Buffer + off, "lpBuffers is NULL\r\n\t");
	}
	if (*lpNumberOfBytesRecvd) {
		off += wsprintf (Buffer + off, "lpNumberOfBytesRecvd: %d (0x%lx)\r\n\t",
			**lpNumberOfBytesRecvd, **lpNumberOfBytesRecvd);
	} else {
		off += wsprintf (Buffer + off, "lpNumberOfBytesRecvd is NULL\r\n\t");
	}
	if (*lpFlags) {
		off += wsprintf (Buffer + off, "lpFlags: %d (0x%lx)\r\n\t",
			**lpFlags, **lpFlags);
	} else {
		off += wsprintf (Buffer + off, "lpFlags is NULL\r\n\t");
	}
	if (*lpFrom) {
		off = FormatSockaddrIn ((LPSOCKADDR_IN)*lpFrom, Buffer, off);
		off += wsprintf (Buffer + off, "\t");
	} else {
		off += wsprintf (Buffer + off, "lpFrom is NULL\r\n\t");
	}
	if (*lpFromlen) {
		off += wsprintf (Buffer + off, "lpFromlen: %d (0x%lx)\r\n\t");
	} else {
		off += wsprintf (Buffer + off, "lpFromlen is NULL\r\n\t");
	}
	if (*lpOverlapped) {
		off += wsprintf (Buffer + off, "lpOverlapped.hEvent: %d (0x%lx)\r\n\t",
			(*lpOverlapped)->hEvent, (*lpOverlapped)->hEvent);
	} else {
		off += wsprintf (Buffer + off, "lpOverlapped is NULL\r\n\t");
	}
	if (*lpCompletionRoutine) {
		wsprintf (Buffer + off, "lpCompletionRoutine: 0x%lx\r\n",
			lpCompletionRoutine);
	} else {
		wsprintf (Buffer + off, "lpCompletionRoutine is NULL\r\n");
	}

    wsprintf(Buffer + Index, "WSARecvFrom() %s.\r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAResetEvent(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    BOOL *RetVal = (BOOL *)ReturnValue;
    WSAEVENT *hEvent = va_arg(vl, WSAEVENT *);

	if (PreOrPost) {
		Index += wsprintf(Buffer + Index, "-> WSAResetEvent()\r\n\t");
		wsprintf (Buffer + Index, "hEvent: %d\r\n", *hEvent);
	} else {
		wsprintf(Buffer + Index, "<- WSAResetEvent() returned: %d\r\n",*RetVal);
	}
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSASend(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    LPWSABUF *lpBuffers = va_arg(vl, LPWSABUF *);
    DWORD *dwBufferCount = va_arg(vl, DWORD *);
    LPDWORD *lpNumberOfBytesSent = va_arg(vl, LPDWORD *);
    DWORD *dwFlags = va_arg(vl, DWORD *);
    LPWSAOVERLAPPED *lpOverlapped = va_arg(vl, LPWSAOVERLAPPED *);
    LPWSAOVERLAPPED_COMPLETION_ROUTINE *lpCompletionRoutine = 
      va_arg(vl, LPWSAOVERLAPPED_COMPLETION_ROUTINE *);

    wsprintf(Buffer + Index, "WSASend() %s.\r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSASendDisconnect(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    LPWSABUF *lpOutboundDisconnectData = va_arg(vl, LPWSABUF *);

	if (PreOrPost) {
		Index += wsprintf(Buffer + Index, "-> WSASendDisconnect()\r\n\t");
		wsprintf (Buffer + Index, "socket: %d\r\n\t", *s);
		if (*lpOutboundDisconnectData) {
			Index = FormatWSABuf (*lpOutboundDisconnectData, Buffer, Index, TRUE);
		} else {
			Index += wsprintf (Buffer + Index, 
				"lpOutboundDisconnectData is NULL\r\n");
		}
		wsprintf (Buffer + Index, "\r\n");
	} else {
		Index += wsprintf(Buffer + Index, 
			"<- WSAResetDisconnect() returned: %d\r\n",*RetVal);
	}
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSASendTo(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    SOCKET *s = va_arg(vl, SOCKET *);
    LPWSABUF *lpBuffers = va_arg(vl, LPWSABUF *);
    DWORD *dwBufferCount = va_arg(vl, DWORD *);
    LPDWORD *lpNumberOfBytesSent = va_arg(vl, LPDWORD *);
    DWORD *dwFlags = va_arg(vl, DWORD *);
    struct sockaddr FAR **lpTo = va_arg(vl, struct sockaddr FAR **);
    int *iTolen = va_arg(vl, int *);
    LPWSAOVERLAPPED *lpOverlapped = va_arg(vl, LPWSAOVERLAPPED *);
    LPWSAOVERLAPPED_COMPLETION_ROUTINE *lpCompletionRoutine = 
      va_arg(vl, LPWSAOVERLAPPED_COMPLETION_ROUTINE *);

    wsprintf(Buffer + Index, "WSASendTo() %s.\r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSASetEvent(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    BOOL *RetVal = (BOOL *)ReturnValue;
    WSAEVENT *hEvent = va_arg(vl, WSAEVENT *);

	if (PreOrPost) {
		Index += wsprintf(Buffer + Index, "-> WSASetEvent()\r\n\t");
		wsprintf (Buffer + Index, "hEvent: %d\r\n", *hEvent);
	} else {
		wsprintf(Buffer + Index, "<- WSASetEvent() returned: %d\r\n",*RetVal);
	}
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSASocketA(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    SOCKET *RetVal = (SOCKET *)ReturnValue;
    int *af = va_arg(vl, int *);
    int *type = va_arg(vl, int *);
    int *protocol = va_arg(vl, int *);
    LPWSAPROTOCOL_INFOA *lpProtocolInfo = va_arg(vl, LPWSAPROTOCOL_INFOA *);
    GROUP *g = va_arg(vl, GROUP *);
    DWORD *dwFlags = va_arg(vl, DWORD *);

	int off = Index;
	if (bOutputEnabled && (dDebugFlags & DT_SOCKET)) {
		if (PreOrPost) {
			off += wsprintf(Buffer + off, 
				"-> WSASocket()\r\n\t");
			off += wsprintf(Buffer + off,"af: %d\r\n\t", *af);
			off += wsprintf(Buffer + off,"type: %d\r\n\t", *type);
			off += wsprintf(Buffer + off,"protocol: %d\r\n\t", *protocol);
			if (*lpProtocolInfo) {
				off = FormatProtocolInfo(*lpProtocolInfo, Buffer, off);
				off =+ wsprintf (Buffer + off, "\t");
			} else {
				off += wsprintf(Buffer + off, "lpProtocolInfo is NULL\r\n\t");
			}
			off += wsprintf(Buffer + off,"group: %d\r\n\t", *g); 
			off += wsprintf(Buffer + off,"dwFlags: %#lx\r\n",
				*g, *dwFlags);
		} else {
			Index += wsprintf(Buffer + Index, 
				"<- WSASocket() returned = %d\r\n",*RetVal);
		}
		DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    }
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSASocketW(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    SOCKET *RetVal = (SOCKET *)ReturnValue;
    int *af = va_arg(vl, int *);
    int *type = va_arg(vl, int *);
    int *protocol = va_arg(vl, int *);
    LPWSAPROTOCOL_INFOW *lpProtocolInfo = va_arg(vl, LPWSAPROTOCOL_INFOW *);
    GROUP *g = va_arg(vl, GROUP *);
    DWORD *dwFlags = va_arg(vl, DWORD *);

    wsprintf(Buffer + Index, "WSASocketW() %s.\r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAWaitForMultipleEvents(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    DWORD *RetVal = (DWORD *)ReturnValue;
    DWORD *cEvents = va_arg(vl, DWORD *);
    WSAEVENT FAR **lphEvents = va_arg(vl, WSAEVENT FAR **);
    BOOL *fWaitAll = va_arg(vl, BOOL *);
    DWORD *dwTimeout = va_arg(vl, DWORD *);
    BOOL *fAlertable = va_arg(vl, BOOL *);
	int off = Index;
	DWORD i;

	if (PreOrPost) {
		off += wsprintf(Buffer + off, 
			"-> WSAWaitForMultipleEvents()\r\n\t");
		off += wsprintf (Buffer + off, "cEvents: %d\r\n\t", *cEvents);
		if (*lphEvents) {
			for (i=0; i<*cEvents; i++) {
				off += wsprintf (Buffer + off,
					"\thEvent[%d]: %d\r\n\t", i, *(lphEvents+i));
			}
		} else {
			off += wsprintf(Buffer + off, "lphEvents is NULL\r\n\t");
		}
		off += wsprintf(Buffer + off, "fWaitAll: %s\r\n\t",
			*fWaitAll ? "TRUE" : "FALSE");
		off += wsprintf(Buffer + off, "dwTimeout: %d (0x%lx)\r\n\t", 
			*dwTimeout, *dwTimeout);
		wsprintf(Buffer + off, "fAlertable: %s\r\n",
			*fAlertable ? "TRUE" : "FALSE");
	} else {
		wsprintf(Buffer + off,
			"<- WSAWaitForMultipleEvents() returned: %d\r\n", *RetVal);
	}
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAAddressToStringA(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPSOCKADDR *lpsaAddress = va_arg(vl, LPSOCKADDR *); 
    DWORD *dwAddressLength= va_arg(vl, DWORD *);
    LPWSAPROTOCOL_INFOA *lpProtocolInfo= va_arg(vl, LPWSAPROTOCOL_INFOA *);
    LPSTR *lpszAddressString= va_arg(vl, LPSTR *);
    LPDWORD *lpdwAddressStringLength= va_arg(vl, LPDWORD *);
      
	int off = Index;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSAAddressToString()\r\n\t");
	} else {
		off += wsprintf(Buffer + off,
			"<- WSAAddressToString() returned: %d\r\n\t", *RetVal);
	}

	if (*lpsaAddress) {
		if ((*lpsaAddress)->sa_family == AF_INET) {
			off = FormatSockaddrIn ((sockaddr_in*)*lpsaAddress, Buffer, off);
			off += wsprintf (Buffer + off, "\t");
		} else {
			off += wsprintf (Buffer + off, 
				"lpsaAddress has an unknown address family (%d)\r\n\t",
				(*lpsaAddress)->sa_family);
		}
	} else {
		off += wsprintf (Buffer + off, "lpsaAddress is NULL\r\n\t");
	}
 	off += wsprintf (Buffer + off, "dwAddressLength: %d (0x%lx)\r\n\t", 
		*dwAddressLength, *dwAddressLength);
	if (*lpProtocolInfo) {
		off = FormatProtocolInfo (*lpProtocolInfo, Buffer, off);
	} else {
		off += wsprintf(Buffer + off, "lpProtocolInfo is NULL\r\n\t");
	}
	if (*lpszAddressString) {
		if (!PreOrPost) {	// only print string upon return
			off += wsprintf (Buffer + off, "lpszAddressString: %s\r\n\t",
				*lpszAddressString);
		}
	} else {
		off += wsprintf (Buffer + off, "lpszAddressString is NULL\r\n\t");
	}
	if (*lpdwAddressStringLength) {
 		off += wsprintf (Buffer + off, 
			"lpdwAddressStringLength: %d (0x%lx)\r\n", 
			**lpdwAddressStringLength, **lpdwAddressStringLength);
	} else {
		off += wsprintf (Buffer + off,
			"lpdwAddressStringLength is NULL\r\n");
	}

	DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}

BOOL CALLBACK
DTHandler_WSAAddressToStringW(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPSOCKADDR *lpsaAddress = va_arg(vl, LPSOCKADDR *); 
    DWORD *dwAddressLength= va_arg(vl, DWORD *);
    LPWSAPROTOCOL_INFOW *lpProtocolInfo= va_arg(vl, LPWSAPROTOCOL_INFOW *);
    LPWSTR *lpszAddressString= va_arg(vl, LPWSTR *);
    LPDWORD *lpdwAddressStringLength= va_arg(vl, LPDWORD *);
      
    wsprintf(Buffer + Index, "WSAAddressToStringW() %s.\r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}

BOOL CALLBACK
DTHandler_WSAStringToAddressA(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPSTR *AddressString = va_arg(vl, LPSTR*);
    INT *AddressFamily = va_arg(vl, INT *);
    LPWSAPROTOCOL_INFOA *lpProtocolInfo = va_arg(vl, LPWSAPROTOCOL_INFOA*);
    LPSOCKADDR *lpAddress = va_arg(vl, LPSOCKADDR*);
    LPINT *lpAddressLength = va_arg(vl, LPINT*);
    
	int off = Index;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSAStringToAddress()\r\n\t");
	} else {
		off += wsprintf(Buffer + off,
			"<- WSAStringToAddress() returned: %d\r\n\t", *RetVal);
	}

	if (*AddressString) {
		off += wsprintf (Buffer + off, "AddressString: %s\r\n\t",
			*AddressString);
	} else {
		off += wsprintf (Buffer + off, "AddressString is NULL\r\n\t");
	}
	off += wsprintf (Buffer + off, "AddressFamily: %d (0x%lx)\r\n\t", 
		*AddressFamily, *AddressFamily);
	if (*lpProtocolInfo) {
		off = FormatProtocolInfo (*lpProtocolInfo, Buffer, off);
		off += wsprintf (Buffer + off, "\t");
	} else {
		off += wsprintf(Buffer + off, "lpProtocolInfo is NULL\r\n\t");
	}
	if (*lpAddress) {
		if ((*lpAddress)->sa_family == AF_INET) {
			off = FormatSockaddrIn ((sockaddr_in*)*lpAddress, Buffer, off);
		} else {
			off += wsprintf (Buffer + off, 
				"lpAddress has an unknown address family (%d)\r\n\t",
				(*lpAddress)->sa_family);
		}
	} else {
		off += wsprintf (Buffer + off, "lpAddress is NULL\r\n\t");
	}
	if (*lpAddressLength) {
 		off += wsprintf (Buffer + off, 
			"lpAddressLength: %d (0x%lx)\r\n", 
			**lpAddressLength, **lpAddressLength);
	} else {
		off += wsprintf (Buffer + off, "lpAddressLength is NULL\r\n");
	}

    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAStringToAddressW(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPWSTR *AddressString = va_arg(vl, LPWSTR*);
    INT *AddressFamily = va_arg(vl, INT *);
    LPWSAPROTOCOL_INFOW *lpProtocolInfo = va_arg(vl, LPWSAPROTOCOL_INFOW*);
    LPSOCKADDR *lpAddress = va_arg(vl, LPSOCKADDR*);
    LPINT *lpAddressLength = va_arg(vl, LPINT*);
    
    wsprintf(Buffer + Index, "WSAStringToAddressW() %s.\r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSALookupServiceBeginA(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPWSAQUERYSETA *lpqsRestrictions = va_arg(vl, LPWSAQUERYSETA*);
    DWORD *dwControlFlags = va_arg(vl, DWORD*);
    LPHANDLE *lphLookup  = va_arg(vl, LPHANDLE*);
 
	int off = Index;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSALookupServiceBegin()\r\n\t");
	} else {
		off += wsprintf(Buffer + off,
			"<- WSALookupServiceBegin() returned: %d\r\n\t", *RetVal);
	}
	if (*lpqsRestrictions) {
		off += wsprintf (Buffer + off, "lpqsRestrictions: \r\n\t");
		off = FormatQuerySet (*lpqsRestrictions, Buffer, off, "\t\t");
		off += wsprintf (Buffer + off, "\t");
	} else {
		off += wsprintf (Buffer + off, "lpqsRestrictions is NULL\r\n\t");
	}
	off += wsprintf (Buffer + off, "dwControlFlags: %d (0x%lx)\r\n\t",
		*dwControlFlags, *dwControlFlags);
	if (*lphLookup) {
		off += wsprintf (Buffer + off, "lphLookup: %d (0x%lx)\r\n",
			**lphLookup, **lphLookup);
	} else {
		off += wsprintf (Buffer + off, "lphLookup is NULL\r\n");
	}
 
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSALookupServiceBeginW(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPWSAQUERYSETW *lpqsRestrictions = va_arg(vl, LPWSAQUERYSETW*);
    DWORD *dwControlFlags = va_arg(vl, DWORD*);
    LPHANDLE *lphLookup  = va_arg(vl, LPHANDLE*);
    
    wsprintf(Buffer + Index, "WSALookupServiceBeginW() %s.\r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSALookupServiceNextA(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    HANDLE *hLookup = va_arg(vl, HANDLE*);
    DWORD *dwControlFlags = va_arg(vl, DWORD*);
    LPDWORD *lpdwBufferLength = va_arg(vl, LPDWORD*);
    LPWSAQUERYSETA *lpqsResults = va_arg(vl, LPWSAQUERYSETA*);
 
	int off = Index;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSALookupServiceNext()\r\n\t");
	} else {
		off += wsprintf(Buffer + off,
			"<- WSALookupServiceNext() returned: %d\r\n\t", *RetVal);
	}
	off += wsprintf (Buffer + off, "lphLookup: %d (0x%lx)\r\n\t",
		*hLookup, *hLookup);
	off += wsprintf (Buffer + off, "dwControlFlags: %d (0x%lx)\r\n\t",
		*dwControlFlags, *dwControlFlags);
	if (*lpdwBufferLength) {
		off += wsprintf (Buffer + off, "lpdwBufferLength: %d (0x%lx)\r\n\t",
			**lpdwBufferLength, **lpdwBufferLength);
	} else {
		off += wsprintf (Buffer + off, "lpdwBufferLength is NULL\r\n\t");
	}
 	if (*lpqsResults) {
		off += wsprintf (Buffer + off, "lpqsResults: \r\n\t");
		off = FormatQuerySet (*lpqsResults, Buffer, off, "\t\t");
		off += wsprintf (Buffer + off, "\t");
	} else {
		off += wsprintf (Buffer + off, "lpqsResults is NULL\r\n");
	}

    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSALookupServiceNextW(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    HANDLE *hLookup = va_arg(vl, HANDLE*);
    DWORD *dwControlFlags = va_arg(vl, DWORD*);
    LPDWORD *lpdwBufferLength = va_arg(vl, LPDWORD*);
    LPWSAQUERYSETW *lpqsResults = va_arg(vl, LPWSAQUERYSETW*);
 
    wsprintf(Buffer + Index, "WSALookupServiceNextW() %s.\r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSALookupServiceEnd(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    HANDLE *hLookup = va_arg(vl, HANDLE*);
    
	int off = Index;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSALookupServiceEnd()\r\n\t");
	} else {
		off += wsprintf(Buffer + off,
			"<- WSALookupServiceEnd() returned: %d\r\n\t", *RetVal);
	}
	off += wsprintf (Buffer + off, "hLookup: %d (0x%lx)\r\n",
		*hLookup, *hLookup);

    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAInstallServiceClassA(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPWSASERVICECLASSINFOA *lpServiceClassInfo =
        va_arg(vl, LPWSASERVICECLASSINFOA*);
 
	int off = Index;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSAInstallServiceClass()\r\n\t");
	} else {
		off += wsprintf(Buffer + off,
			"<- WSAInstallServiceClass() returned: %d\r\n\t", *RetVal);
	}
	if (*lpServiceClassInfo) {
		if (!IsBadReadPtr(*lpServiceClassInfo, sizeof(WSASERVICECLASSINFOA))) {
			off += wsprintf (Buffer + off, "lpServiceClassInfo: \r\n\t");
			off = FormatServiceClassInfo (*lpServiceClassInfo, Buffer, off, "\t\t");
		} else {
			off += wsprintf (Buffer + off, "lpServiceClassInfo POINTER IS INVALID\r\n");
		}
	} else {
		off += wsprintf (Buffer + off, "lpServiceClassInfo is NULL\r\n");
	}

    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAInstallServiceClassW(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPWSASERVICECLASSINFOW *lpServiceClassInfo =
        va_arg(vl, LPWSASERVICECLASSINFOW*);
 
    wsprintf(Buffer + Index, "WSAInstallServiceClassW() %s.\r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSASetServiceA(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPWSAQUERYSETA *lpqsRegInfo = va_arg(vl, LPWSAQUERYSETA*);
    WSAESETSERVICEOP *essOperation = va_arg(vl, WSAESETSERVICEOP*);
    DWORD *dwControlFlags = va_arg(vl, DWORD*);
 
	int off = Index;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSASetService()\r\n\t");
	} else {
		off += wsprintf(Buffer + off,
			"<- WSASetService() returned: %d\r\n\t", *RetVal);
	}
	if (*lpqsRegInfo) {
		off += wsprintf (Buffer + off, "lpqsRegInfo: \r\n\t");
		off = FormatQuerySet (*lpqsRegInfo, Buffer, off, "\t\t");
		off += wsprintf (Buffer + off, "\t");
	} else {
		off += wsprintf (Buffer + off, "lpqsRegInfo is NULL\r\n\t");
	}
	off += wsprintf (Buffer + off, "essOperation: %s (0x%lx)\r\n\t",
		setservops[*essOperation], *essOperation);
	off += wsprintf (Buffer + off, "dwControlFlags: %d (0x%lx)\r\n",
		*dwControlFlags, *dwControlFlags);

    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSASetServiceW(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPWSAQUERYSETW *lpqsRegInfo = va_arg(vl, LPWSAQUERYSETW*);
    WSAESETSERVICEOP *essOperation = va_arg(vl, WSAESETSERVICEOP*);
    DWORD *dwControlFlags = va_arg(vl, DWORD*);
 
    wsprintf(Buffer + Index, "WSASetServiceW() %s.\r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSARemoveServiceClass(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPGUID *lpServiceClassId = va_arg(vl, LPGUID*);

	int off = Index;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSARemoveServiceClass()\r\n\t");
	} else {
		off += wsprintf(Buffer + off,
			"<- WSARemoveServiceClass() returned: %d\r\n\t", *RetVal);
	}
	if (*lpServiceClassId) {
		if (!IsBadReadPtr (*lpServiceClassId, sizeof(GUID))) {
			off += wsprintf (Buffer + off, "lpServiceClassId: \r\n\t");
			off = FormatGUID (*lpServiceClassId, Buffer, off, "\t\t");
		} else {
			off += wsprintf (Buffer + off, 
				"lpServiceClassId POINTER IS INVALID\r\n");
		}
	} else {
		off += wsprintf (Buffer + off, "lpServiceClassId is NULL\r\n");
	}
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAGetServiceClassInfoA(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPGUID *lpProviderId = va_arg(vl, LPGUID*);
    LPGUID *lpServiceClassId = va_arg(vl, LPGUID*);
    LPDWORD *lpdwBufSize = va_arg(vl, LPDWORD*);
    LPWSASERVICECLASSINFOA *lpServiceClassInfo =
        va_arg(vl, LPWSASERVICECLASSINFOA*);

	int off = Index;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSAGetServiceClassInfo()\r\n\t");
	} else {
		off += wsprintf(Buffer + off,
			"<- WSAGetServiceClassInfo() returned: %d\r\n\t", *RetVal);
	}
	if (*lpProviderId) {
		if (!IsBadReadPtr (*lpProviderId, sizeof(GUID))) {
			off += wsprintf (Buffer + off, "lpProviderId: \r\n\t");
			off = FormatGUID (*lpProviderId, Buffer, off, "\t\t");
		} else {
			off += wsprintf (Buffer + off, 
				"lpProviderId POINTER IS INVALID\r\n");
		}
	} else {
		off += wsprintf (Buffer + off, "lpProviderId is NULL\r\n");
	}
	if (*lpServiceClassId) {
		if (!IsBadReadPtr (*lpServiceClassId, sizeof(GUID))) {
			off += wsprintf (Buffer + off, "lpServiceClassId: \r\n\t");
			off = FormatGUID (*lpServiceClassId, Buffer, off, "\t\t");
		} else {
			off += wsprintf (Buffer + off, 
				"lpServiceClassId POINTER IS INVALID\r\n");
		}
	} else {
		off += wsprintf (Buffer + off, "lpServiceClassId is NULL\r\n");
	}
	if (*lpdwBufSize) {
		if (!IsBadReadPtr(*lpdwBufSize, sizeof(DWORD))) {
			off += wsprintf (Buffer + off, "lpdwBufSize: %d (0x%lx)\r\n",
				**lpdwBufSize, **lpdwBufSize);
		} else {
			off += wsprintf (Buffer + off, "lpdwBufSize POINTER IS INVALID\r\n");
		}
	} else {
		off += wsprintf (Buffer + off, "lpdwBufSize is NULL\r\n");
	}
	if (*lpServiceClassInfo) {
		if (!IsBadReadPtr(*lpServiceClassInfo, sizeof(WSASERVICECLASSINFOA))) {
			off += wsprintf (Buffer + off, "lpServiceClassInfo: \r\n\t");
			off = FormatServiceClassInfo (*lpServiceClassInfo, Buffer, off, "\t\t");
		} else {
			off += wsprintf (Buffer + off, "lpServiceClassInfo POINTER IS INVALID\r\n");
		}
	} else {
		off += wsprintf (Buffer + off, "lpServiceClassInfo is NULL\r\n");
	}

    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}



BOOL CALLBACK
DTHandler_WSAGetServiceClassInfoW(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPGUID *lpProviderId = va_arg(vl, LPGUID*);
    LPGUID *lpServiceClassId = va_arg(vl, LPGUID*);
    LPDWORD *lpdwBufSize = va_arg(vl, LPDWORD*);
    LPWSASERVICECLASSINFOW *lpServiceClassInfo =
        va_arg(vl, LPWSASERVICECLASSINFOW*);

    wsprintf(Buffer + Index, "WSAGetServiceClassInfoW() %s.\r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAEnumNameSpaceProvidersA(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPDWORD *lpdwBufferLength = va_arg(vl, LPDWORD*);
    LPWSANAMESPACE_INFOA *lpnspBuffer = va_arg(vl,  LPWSANAMESPACE_INFOA*);
 
	int off = Index;
	if (PreOrPost) {
		off += wsprintf(Buffer + off, "-> WSAEnumNameSpaceProviders()\r\n\t");
	} else {
		off += wsprintf(Buffer + off,
			"<- WSAEnumNameSpaceProviders() returned: %d\r\n\t", *RetVal);
	}
	if (*lpdwBufferLength) {
		if (!IsBadReadPtr(*lpdwBufferLength, sizeof(DWORD))) {
			off += wsprintf (Buffer + off, "lpdwBufferLength: %d (0x%lx)\r\n",
				**lpdwBufferLength, **lpdwBufferLength);
		} else {
			off += wsprintf (Buffer + off, "lpdwBufferLength POINTER IS INVALID\r\n");
		}
	} else {
		off += wsprintf (Buffer + off, "lpdwBufferLength is NULL\r\n");
	}
	//
	// bq note: Need to go through the entire NSP Buffer for all NSP info structs
	//
	if (*lpnspBuffer) {
		if (!IsBadReadPtr (*lpnspBuffer, sizeof(GUID))) {
			off += wsprintf (Buffer + off, "lpnspBuffer: \r\n\t");
			off = FormatNameSpaceInfo (*lpnspBuffer, Buffer, off, "\t\t");
		} else {
			off += wsprintf (Buffer + off, 
				"lpnspBuffer POINTER IS INVALID\r\n");
		}
	} else {
		off += wsprintf (Buffer + off, "lpnspBuffer is NULL\r\n");
	}

    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAEnumNameSpaceProvidersW(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPDWORD *lpdwBufferLength = va_arg(vl, LPDWORD*);
    LPWSANAMESPACE_INFOW *lpnspBuffer = va_arg(vl,  LPWSANAMESPACE_INFOW*);
 
    wsprintf(Buffer + Index, "WSAEnumNameSpaceProvidersW() %s.\r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAGetServiceClassNameByClassIdA(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPGUID *lpServiceClassId = va_arg(vl, LPGUID*);
    LPSTR *lpszServiceClassName = va_arg(vl, LPSTR*);
    LPDWORD *lpdwBufferLength = va_arg(vl, LPDWORD*); 
     
    wsprintf(Buffer + Index, "WSAGetServiceClassNameByClassIdA() %s.\r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}


BOOL CALLBACK
DTHandler_WSAGetServiceClassNameByClassIdW(
    IN     va_list vl,
    IN OUT LPVOID  ReturnValue,
    IN     LPSTR   LibraryName,
    OUT    char    *Buffer,
    IN     int     Index,
    IN     int     BufLen,
    IN     BOOL    PreOrPost)
{
    int *RetVal = (int *)ReturnValue;
    LPGUID *lpServiceClassId = va_arg(vl, LPGUID*);
    LPWSTR *lpszServiceClassName = va_arg(vl, LPWSTR*);
    LPDWORD *lpdwBufferLength = va_arg(vl, LPDWORD*); 
     
    wsprintf(Buffer + Index, "WSAGetServiceClassNameByClassIdW() %s.\r\n",
             PreOrPost ? "->" : "<-");
    DTTextOut(DebugWindow, LogFileHandle, Buffer, OutputStyle);
    return(FALSE);
}





